import os
import os.path
import shutil
import textwrap
import unittest

import pyperformance
from pyperformance import tests


class FullStackTests(tests.Functional, unittest.TestCase):
    maxDiff = 80 * 100

    @classmethod
    def setUpClass(cls):
        # pyperformance must be installed in order to run,
        # so we make sure it is.
        cls.ensure_venv()
        cls.ensure_pyperformance()
        super().setUpClass()

    @classmethod
    def ensure_pyperformance(cls):
        ec, stdout, _ = cls.run_python(
            os.path.join(tests.DATA_DIR, "find-pyperformance.py"),
            capture="stdout",
            onfail="raise",
            verbose=False,
        )
        assert ec == 0, ec
        stdout = stdout.strip()
        if stdout.strip():
            # It is already installed.
            return

        print("#" * 40)
        print("# installing pyperformance into the venv")
        print("#" * 40)
        print()

        # Install it.
        reporoot = os.path.dirname(pyperformance.PKG_ROOT)
        # XXX Ignore the output (and optionally log it).
        ec, _, _ = cls.run_pip("install", "--editable", reporoot)
        assert ec == 0, ec

        # Clean up extraneous files.
        egg_info = "pyperformance.egg-info"
        print(f"(tests) Remove directory {egg_info}", flush=True)
        try:
            shutil.rmtree(egg_info)
        except FileNotFoundError:
            pass
        print()

        print("#" * 40)
        print("# DONE: installing pyperformance into the venv")
        print("#" * 40)
        print()

    def run_pyperformance(
        self,
        cmd,
        *args,
        exitcode=0,
        capture="both",
        verbose=True,
    ):
        ec, stdout, stderr = self.run_module(
            "pyperformance",
            cmd,
            *args,
            capture=capture,
            onfail=None,
            verbose=verbose,
        )
        if exitcode is True:
            self.assertGreater(ec, 0, repr(stdout))
        else:
            self.assertEqual(ec, exitcode, repr(stdout))
        if stdout:
            stdout = stdout.rstrip()
        return stdout

    ###################################
    # info

    def test_list(self):
        # XXX Capture and check the output.
        self.run_pyperformance("list", capture=None)

    def test_list_groups(self):
        # XXX Capture and check the output.
        self.run_pyperformance("list_groups", capture=None)

    ###################################
    # venv

    def test_venv(self):
        # XXX Capture and check the output.
        root = self.resolve_tmp("venv", unique=True)

        def div():
            print()
            print("---")
            print()

        def expect_success(*args):
            self.run_pyperformance(
                *args,
                capture=None,
            )

        def expect_failure(*args):
            self.run_pyperformance(
                *args,
                capture=None,
                exitcode=1,
            )

        # It doesn't exist yet.
        expect_success("venv", "show", "--venv", root)
        div()
        # It gets created.
        expect_success("venv", "create", "--venv", root)
        div()
        expect_success("venv", "show", "--venv", root)
        div()
        # It alraedy exists.
        expect_failure("venv", "create", "--venv", root)
        div()
        expect_success("venv", "show", "--venv", root)
        div()
        # It gets re-created.
        expect_success("venv", "recreate", "--venv", root)
        div()
        expect_success("venv", "show", "--venv", root)
        div()
        # It get deleted.
        expect_success("venv", "remove", "--venv", root)
        div()
        expect_success("venv", "show", "--venv", root)

    ###################################
    # run

    def test_run_and_show(self):
        filename = self.resolve_tmp("bench.json")

        # -b all: check that *all* benchmark work
        #
        # --debug-single-value: benchmark results don't matter, we only
        # check that running benchmarks don't fail.
        # XXX Capture and check the output.
        self.run_pyperformance(
            "run",
            "-b",
            "all",
            "--debug-single-value",
            "-o",
            filename,
            capture=None,
        )

        # Display slowest benchmarks
        # XXX Capture and check the output.
        self.run_module("pyperf", "slowest", filename)

    def test_run_test_benchmarks(self):
        # Run the benchmarks that exist only for testing
        # in pyperformance/tests/data
        filename = self.resolve_tmp("bench-test.json")

        self.run_pyperformance(
            "run",
            "--manifest",
            os.path.join(tests.DATA_DIR, "MANIFEST"),
            "-b",
            "all",
            "-o",
            filename,
            capture=None,
        )

    def test_run_with_hook(self):
        # We expect this to fail, since pystats requires a special build of Python
        filename = self.resolve_tmp("bench-test-hook.json")

        stdout = self.run_pyperformance(
            "run",
            "--manifest",
            os.path.join(tests.DATA_DIR, "MANIFEST"),
            "-b",
            "all",
            "-o",
            filename,
            "--hook",
            "pystats",
            exitcode=1,
            capture="combined",
        )

        self.assertIn(
            "Can not collect pystats because python was not built with --enable-pystats",
            stdout,
        )

    ###################################
    # compile

    def ensure_cpython_repo(self, reporoot=None):
        if not reporoot:
            reporoot = os.environ.get("PYPERFORMANCE_TESTS_CPYTHON")
            if not reporoot:
                reporoot = os.path.join(tests.DATA_DIR, "cpython")
        for markerfile in [
            os.path.join(reporoot, ".git"),
            os.path.join(reporoot, "Python/ceval.c"),
        ]:
            if not os.path.exists(markerfile):
                break
        else:
            return reporoot
        # Clone the repo.
        print("#" * 40)
        print("# cloning the cpython repo")
        print("#" * 40)
        print()
        tests.run_cmd(
            shutil.which("git"),
            "clone",
            "https://github.com/python/cpython",
            reporoot,
        )
        print("#" * 40)
        print("# DONE: cloning the cpython repo")
        print("#" * 40)
        print()
        return reporoot

    def create_compile_config(
        self,
        *revisions,
        outdir=None,
        fast=True,
        upload=None,
    ):
        if not outdir:
            outdir = self.resolve_tmp("compile-cmd-outdir", unique=True)
        cpython = self.ensure_cpython_repo()
        text = textwrap.dedent(f"""
            [config]
            json_dir = {outdir}
            debug = {fast}

            [scm]
            repo_dir = {cpython}
            update = False
            git_remote = remotes/origin

            [compile]
            bench_dir = {outdir}
            lto = {not fast}
            pgo = {not fast}
            jit = no
            install = True

            [run_benchmark]
            system_tune = False
            upload = False

            [upload]
            url = {upload}
            """)
        if revisions:
            text += "".join(
                line + os.linesep
                for line in [
                    "",
                    "[compile_all_revisions]",
                    *(f"{r} =" for r in revisions),
                ]
            )
        cfgfile = os.path.join(outdir, "compile.ini")
        print(f"(writing config file to {cfgfile})")
        os.makedirs(outdir, exist_ok=True)
        with open(cfgfile, "w", encoding="utf-8") as outfile:
            outfile.write(text)
        return cfgfile

    @tests.CPYTHON_ONLY
    @tests.NON_WINDOWS_ONLY
    @tests.SLOW
    def test_compile(self):
        cfgfile = self.create_compile_config()
        revision = "a58ebcc701dd"  # tag: v3.10.2

        # XXX Capture and check the output.
        self.run_pyperformance(
            "compile",
            cfgfile,
            revision,
            capture=None,
        )

    @tests.CPYTHON_ONLY
    @tests.NON_WINDOWS_ONLY
    @tests.SLOW
    def test_compile_all(self):
        rev1 = "2cd268a3a934"  # tag: v3.10.1
        rev2 = "a58ebcc701dd"  # tag: v3.10.2
        cfgfile = self.create_compile_config(rev1, rev2)

        # XXX Capture and check the output.
        self.run_pyperformance(
            "compile_all",
            cfgfile,
            capture=None,
        )

    @tests.CPYTHON_ONLY
    @tests.NON_WINDOWS_ONLY
    @unittest.expectedFailure
    def test_upload(self):
        url = "<bogus>"
        cfgfile = self.create_compile_config(upload=url)
        resfile = os.path.join(tests.DATA_DIR, "py36.json")

        # XXX Capture and check the output.
        self.run_pyperformance(
            "upload",
            cfgfile,
            resfile,
            capture=None,
        )

    ###################################
    # show

    def test_show(self):
        for filename in (
            os.path.join(tests.DATA_DIR, "py36.json"),
            os.path.join(tests.DATA_DIR, "mem1.json"),
        ):
            with self.subTest(filename):
                # XXX Capture and check the output.
                self.run_pyperformance("show", filename, capture=None)

    ###################################
    # compare

    def compare(self, *args, exitcode=0, dataset="py", file2="py38.json", **kw):
        if dataset == "mem":
            file1 = "mem1.json"
            file2 = "mem2.json"
        else:
            file1 = "py36.json"
        marker = file1

        stdout = self.run_pyperformance(
            "compare",
            os.path.join(tests.DATA_DIR, file1),
            os.path.join(tests.DATA_DIR, file2),
            *args,
            exitcode=exitcode,
            verbose=False,
        )
        if marker in stdout:
            stdout = stdout[stdout.index(marker) :]
        return stdout + "\n"

    def test_compare(self):
        stdout = self.compare()
        self.assertEqual(
            stdout,
            textwrap.dedent("""
            py36.json
            =========

            Performance version: 1.0.1
            Python version: 3.6.10 (64-bit)
            Report on Linux-5.5.9-200.fc31.x86_64-x86_64-with-fedora-31-Thirty_One
            Number of logical CPUs: 8
            Start date: 2020-03-26 15:50:39.816020
            End date: 2020-03-26 15:50:56.406559

            py38.json
            =========

            Performance version: 1.0.1
            Python version: 3.8.2 (64-bit)
            Report on Linux-5.5.9-200.fc31.x86_64-x86_64-with-glibc2.2.5
            Number of logical CPUs: 8
            Start date: 2020-03-26 15:54:12.331569
            End date: 2020-03-26 15:54:23.900355

            ### telco ###
            Mean +- std dev: 10.7 ms +- 0.5 ms -> 7.2 ms +- 0.3 ms: 1.49x faster
            Significant (t=44.97)
        """).lstrip(),
        )

    def test_compare_wrong_version(self):
        stdout = self.compare(file2="py3_performance03.json", exitcode=1)
        self.assertEqual(
            stdout,
            textwrap.dedent("""
            py36.json
            =========

            Performance version: 1.0.1
            Python version: 3.6.10 (64-bit)
            Report on Linux-5.5.9-200.fc31.x86_64-x86_64-with-fedora-31-Thirty_One
            Number of logical CPUs: 8
            Start date: 2020-03-26 15:50:39.816020
            End date: 2020-03-26 15:50:56.406559

            py3_performance03.json
            ======================

            Performance version: 0.3


            Skipped 1 benchmarks only in py36.json: telco

            Skipped 1 benchmarks only in py3_performance03.json: call_simple
            ERROR: Performance versions are different (1.0.1 != 0.3)
            """).lstrip(),
        )

    def test_compare_single_value(self):
        stdout = self.compare(dataset="mem")
        self.assertEqual(
            stdout,
            textwrap.dedent("""
            mem1.json
            =========

            Performance version: 0.2

            mem2.json
            =========

            Performance version: 0.2

            ### call_simple ###
            7896.0 KiB -> 7900.0 KiB: 1.00x larger
        """).lstrip(),
        )

    def test_compare_csv(self):
        expected = textwrap.dedent("""
            Benchmark,Base,Changed
            telco,0.01073,0.00722
            """).lstrip()
        filename = self.resolve_tmp("outfile.csv", unique=True)
        with tests.CleanupFile(filename):
            self.compare("--csv", filename)
            with open(filename, "r", encoding="utf-8") as infile:
                csv = infile.read()

        self.assertEqual(csv, expected)

    def test_compare_table(self):
        stdout = self.compare("-O", "table")
        self.assertEqual(
            stdout,
            textwrap.dedent("""
            py36.json
            =========

            Performance version: 1.0.1
            Python version: 3.6.10 (64-bit)
            Report on Linux-5.5.9-200.fc31.x86_64-x86_64-with-fedora-31-Thirty_One
            Number of logical CPUs: 8
            Start date: 2020-03-26 15:50:39.816020
            End date: 2020-03-26 15:50:56.406559

            py38.json
            =========

            Performance version: 1.0.1
            Python version: 3.8.2 (64-bit)
            Report on Linux-5.5.9-200.fc31.x86_64-x86_64-with-glibc2.2.5
            Number of logical CPUs: 8
            Start date: 2020-03-26 15:54:12.331569
            End date: 2020-03-26 15:54:23.900355

            +-----------+-----------+-----------+--------------+-----------------------+
            | Benchmark | py36.json | py38.json | Change       | Significance          |
            +===========+===========+===========+==============+=======================+
            | telco     | 10.7 ms   | 7.22 ms   | 1.49x faster | Significant (t=44.97) |
            +-----------+-----------+-----------+--------------+-----------------------+
        """).lstrip(),
        )

    def test_compare_table_single_value(self):
        stdout = self.compare("-O", "table", dataset="mem")
        self.assertEqual(
            stdout,
            textwrap.dedent("""
            mem1.json
            =========

            Performance version: 0.2

            mem2.json
            =========

            Performance version: 0.2

            +-------------+------------+------------+--------------+------------------------------------------+
            | Benchmark   | mem1.json  | mem2.json  | Change       | Significance                             |
            +=============+============+============+==============+==========================================+
            | call_simple | 7896.0 KiB | 7900.0 KiB | 1.00x larger | (benchmark only contains a single value) |
            +-------------+------------+------------+--------------+------------------------------------------+
        """).lstrip(),
        )


if __name__ == "__main__":
    unittest.main()



import pyperf

from sympy import expand, symbols, integrate, tan, summation
from sympy.core.cache import clear_cache


def bench_expand():
    x, y, z = symbols('x y z')
    expand((1 + x + y + z) ** 20)


def bench_integrate():
    x, y = symbols('x y')
    f = (1 / tan(x)) ** 10
    return integrate(f, x)


def bench_sum():
    x, i = symbols('x i')
    summation(x ** i / i, (i, 1, 400))


def bench_str():
    x, y, z = symbols('x y z')
    str(expand((x + 2 * y + 3 * z) ** 30))


def bench_sympy(loops, func):
    timer = pyperf.perf_counter
    dt = 0

    for _ in range(loops):
        # Don't benchmark clear_cache(), exclude it of the benchmark
        clear_cache()

        t0 = timer()
        func()
        dt += (timer() - t0)

    return dt


BENCHMARKS = ("expand", "integrate", "sum", "str")


def add_cmdline_args(cmd, args):
    if args.benchmark:
        cmd.append(args.benchmark)


if __name__ == "__main__":
    runner = pyperf.Runner(add_cmdline_args=add_cmdline_args)
    runner.metadata['description'] = "SymPy benchmark"
    runner.argparser.add_argument("benchmark", nargs='?',
                                  choices=BENCHMARKS)

    import gc
    gc.disable()

    args = runner.parse_args()
    if args.benchmark:
        benchmarks = (args.benchmark,)
    else:
        benchmarks = BENCHMARKS

    for bench in benchmarks:
        name = 'sympy_%s' % bench
        func = globals()['bench_' + bench]
        runner.bench_time_func(name, bench_sympy, func)



import unittest

from pyperformance import tests


def load_tests(loader, standard_tests, pattern):
    pkgtests = loader.discover(
        start_dir=tests.TESTS_ROOT,
        top_level_dir=tests.TESTS_ROOT,
        pattern=pattern or "test*",
    )
    standard_tests.addTests(pkgtests)
    return standard_tests


if __name__ == "__main__":
    unittest.main()



"""
based on a Java version:
 Based on original version written in BCPL by Dr Martin Richards
 in 1981 at Cambridge University Computer Laboratory, England
 and a C++ version derived from a Smalltalk version written by
 L Peter Deutsch.
 Java version:  Copyright (C) 1995 Sun Microsystems, Inc.
 Translation from C++, Mario Wolczko
 Outer loop added by Alex Jacoby
"""

import pyperf


# Task IDs
I_IDLE = 1
I_WORK = 2
I_HANDLERA = 3
I_HANDLERB = 4
I_DEVA = 5
I_DEVB = 6

# Packet types
K_DEV = 1000
K_WORK = 1001

# Packet

BUFSIZE = 4

BUFSIZE_RANGE = range(BUFSIZE)


class Packet(object):

    def __init__(self, l, i, k):
        self.link = l
        self.ident = i
        self.kind = k
        self.datum = 0
        self.data = [0] * BUFSIZE

    def append_to(self, lst):
        self.link = None
        if lst is None:
            return self
        else:
            p = lst
            next = p.link
            while next is not None:
                p = next
                next = p.link
            p.link = self
            return lst

# Task Records


class TaskRec(object):
    pass


class DeviceTaskRec(TaskRec):

    def __init__(self):
        self.pending = None


class IdleTaskRec(TaskRec):

    def __init__(self):
        self.control = 1
        self.count = 10000


class HandlerTaskRec(TaskRec):

    def __init__(self):
        self.work_in = None
        self.device_in = None

    def workInAdd(self, p):
        self.work_in = p.append_to(self.work_in)
        return self.work_in

    def deviceInAdd(self, p):
        self.device_in = p.append_to(self.device_in)
        return self.device_in


class WorkerTaskRec(TaskRec):

    def __init__(self):
        self.destination = I_HANDLERA
        self.count = 0
# Task


class TaskState(object):

    def __init__(self):
        self.packet_pending = True
        self.task_waiting = False
        self.task_holding = False

    def packetPending(self):
        self.packet_pending = True
        self.task_waiting = False
        self.task_holding = False
        return self

    def waiting(self):
        self.packet_pending = False
        self.task_waiting = True
        self.task_holding = False
        return self

    def running(self):
        self.packet_pending = False
        self.task_waiting = False
        self.task_holding = False
        return self

    def waitingWithPacket(self):
        self.packet_pending = True
        self.task_waiting = True
        self.task_holding = False
        return self

    def isPacketPending(self):
        return self.packet_pending

    def isTaskWaiting(self):
        return self.task_waiting

    def isTaskHolding(self):
        return self.task_holding

    def isTaskHoldingOrWaiting(self):
        return self.task_holding or (not self.packet_pending and self.task_waiting)

    def isWaitingWithPacket(self):
        return self.packet_pending and self.task_waiting and not self.task_holding


tracing = False
layout = 0


def trace(a):
    global layout
    layout -= 1
    if layout <= 0:
        print()
        layout = 50
    print(a, end='')


TASKTABSIZE = 10


class TaskWorkArea(object):

    def __init__(self):
        self.taskTab = [None] * TASKTABSIZE

        self.taskList = None

        self.holdCount = 0
        self.qpktCount = 0


taskWorkArea = TaskWorkArea()


class Task(TaskState):

    def __init__(self, i, p, w, initialState, r):
        self.link = taskWorkArea.taskList
        self.ident = i
        self.priority = p
        self.input = w

        self.packet_pending = initialState.isPacketPending()
        self.task_waiting = initialState.isTaskWaiting()
        self.task_holding = initialState.isTaskHolding()

        self.handle = r

        taskWorkArea.taskList = self
        taskWorkArea.taskTab[i] = self

    def fn(self, pkt, r):
        raise NotImplementedError

    def addPacket(self, p, old):
        if self.input is None:
            self.input = p
            self.packet_pending = True
            if self.priority > old.priority:
                return self
        else:
            p.append_to(self.input)
        return old

    def runTask(self):
        if self.isWaitingWithPacket():
            msg = self.input
            self.input = msg.link
            if self.input is None:
                self.running()
            else:
                self.packetPending()
        else:
            msg = None

        return self.fn(msg, self.handle)

    def waitTask(self):
        self.task_waiting = True
        return self

    def hold(self):
        taskWorkArea.holdCount += 1
        self.task_holding = True
        return self.link

    def release(self, i):
        t = self.findtcb(i)
        t.task_holding = False
        if t.priority > self.priority:
            return t
        else:
            return self

    def qpkt(self, pkt):
        t = self.findtcb(pkt.ident)
        taskWorkArea.qpktCount += 1
        pkt.link = None
        pkt.ident = self.ident
        return t.addPacket(pkt, self)

    def findtcb(self, id):
        t = taskWorkArea.taskTab[id]
        if t is None:
            raise Exception("Bad task id %d" % id)
        return t


# DeviceTask


class DeviceTask(Task):

    def __init__(self, i, p, w, s, r):
        Task.__init__(self, i, p, w, s, r)

    def fn(self, pkt, r):
        d = r
        assert isinstance(d, DeviceTaskRec)
        if pkt is None:
            pkt = d.pending
            if pkt is None:
                return self.waitTask()
            else:
                d.pending = None
                return self.qpkt(pkt)
        else:
            d.pending = pkt
            if tracing:
                trace(pkt.datum)
            return self.hold()


class HandlerTask(Task):

    def __init__(self, i, p, w, s, r):
        Task.__init__(self, i, p, w, s, r)

    def fn(self, pkt, r):
        h = r
        assert isinstance(h, HandlerTaskRec)
        if pkt is not None:
            if pkt.kind == K_WORK:
                h.workInAdd(pkt)
            else:
                h.deviceInAdd(pkt)
        work = h.work_in
        if work is None:
            return self.waitTask()
        count = work.datum
        if count >= BUFSIZE:
            h.work_in = work.link
            return self.qpkt(work)

        dev = h.device_in
        if dev is None:
            return self.waitTask()

        h.device_in = dev.link
        dev.datum = work.data[count]
        work.datum = count + 1
        return self.qpkt(dev)

# IdleTask


class IdleTask(Task):

    def __init__(self, i, p, w, s, r):
        Task.__init__(self, i, 0, None, s, r)

    def fn(self, pkt, r):
        i = r
        assert isinstance(i, IdleTaskRec)
        i.count -= 1
        if i.count == 0:
            return self.hold()
        elif i.control & 1 == 0:
            i.control //= 2
            return self.release(I_DEVA)
        else:
            i.control = i.control // 2 ^ 0xd008
            return self.release(I_DEVB)


# WorkTask


A = ord('A')


class WorkTask(Task):

    def __init__(self, i, p, w, s, r):
        Task.__init__(self, i, p, w, s, r)

    def fn(self, pkt, r):
        w = r
        assert isinstance(w, WorkerTaskRec)
        if pkt is None:
            return self.waitTask()

        if w.destination == I_HANDLERA:
            dest = I_HANDLERB
        else:
            dest = I_HANDLERA

        w.destination = dest
        pkt.ident = dest
        pkt.datum = 0

        for i in BUFSIZE_RANGE:  # range(BUFSIZE)
            w.count += 1
            if w.count > 26:
                w.count = 1
            pkt.data[i] = A + w.count - 1

        return self.qpkt(pkt)


def schedule():
    t = taskWorkArea.taskList
    while t is not None:
        if tracing:
            print("tcb =", t.ident)

        if t.isTaskHoldingOrWaiting():
            t = t.link
        else:
            if tracing:
                trace(chr(ord("0") + t.ident))
            t = t.runTask()


class Richards(object):

    def run(self, iterations):
        for i in range(iterations):
            taskWorkArea.holdCount = 0
            taskWorkArea.qpktCount = 0

            IdleTask(I_IDLE, 1, 10000, TaskState().running(), IdleTaskRec())

            wkq = Packet(None, 0, K_WORK)
            wkq = Packet(wkq, 0, K_WORK)
            WorkTask(I_WORK, 1000, wkq, TaskState(
            ).waitingWithPacket(), WorkerTaskRec())

            wkq = Packet(None, I_DEVA, K_DEV)
            wkq = Packet(wkq, I_DEVA, K_DEV)
            wkq = Packet(wkq, I_DEVA, K_DEV)
            HandlerTask(I_HANDLERA, 2000, wkq, TaskState(
            ).waitingWithPacket(), HandlerTaskRec())

            wkq = Packet(None, I_DEVB, K_DEV)
            wkq = Packet(wkq, I_DEVB, K_DEV)
            wkq = Packet(wkq, I_DEVB, K_DEV)
            HandlerTask(I_HANDLERB, 3000, wkq, TaskState(
            ).waitingWithPacket(), HandlerTaskRec())

            wkq = None
            DeviceTask(I_DEVA, 4000, wkq,
                       TaskState().waiting(), DeviceTaskRec())
            DeviceTask(I_DEVB, 5000, wkq,
                       TaskState().waiting(), DeviceTaskRec())

            schedule()

            if taskWorkArea.holdCount == 9297 and taskWorkArea.qpktCount == 23246:
                pass
            else:
                return False

        return True


if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata['description'] = "The Richards benchmark"

    richard = Richards()
    runner.bench_func('richards', richard.run, 1)



import os
import sys

sys.path.append(os.path.abspath('tools/extensions'))

extensions = [
    'pyspecific',
    'sphinx.ext.extlinks',
]

manpages_url = 'https://manpages.debian.org/{path}'

# General substitutions.
project = 'Python'
copyright = f"2001, Python Software Foundation"

version = release = sys.version.split(" ", 1)[0]

rst_epilog = f"""
.. |python_version_literal| replace:: ``Python {version}``
.. |python_x_dot_y_literal| replace:: ``python{version}``
.. |usr_local_bin_python_x_dot_y_literal| replace:: ``/usr/local/bin/python{version}``
"""

# There are two options for replacing |today|: either, you set today to some
# non-false value, then it is used:
today = ''
# Else, today_fmt is used as the format for a strftime call.
today_fmt = '%B %d, %Y'

# By default, highlight as Python 3.
highlight_language = 'python3'

# Minimum version of sphinx required
needs_sphinx = '6.2.1'

# Create table of contents entries for domain objects (e.g. functions, classes,
# attributes, etc.). Default is True.
toc_object_entries = False

# Disable Docutils smartquotes for several translations
smartquotes_excludes = {
    'languages': ['ja', 'fr', 'zh_TW', 'zh_CN'],
    'builders': ['man', 'text'],
}

# Avoid a warning with Sphinx >= 4.0
root_doc = 'contents'

extlinks = {
    "cve": ("https://cve.mitre.org/cgi-bin/cvename.cgi?name=CVE-%s", "CVE-%s"),
    "cwe": ("https://cwe.mitre.org/data/definitions/%s.html", "CWE-%s"),
    "pypi": ("https://pypi.org/project/%s/", "%s"),
    "source": ('https://github.com/python/cpython/tree/3.13/%s', "%s"),
}
extlinks_detect_hardcoded_links = True



"""
Some graph algorithm benchmarks using networkx

This uses the public domain Amazon data set from the SNAP benchmarks:

    https://snap.stanford.edu/data/amazon0302.html

Choice of benchmarks inspired by Timothy Lin's work here:

    https://www.timlrx.com/blog/benchmark-of-popular-graph-network-packages
"""

import collections
from pathlib import Path

import networkx

import pyperf


DATA_FILE = Path(__file__).parent / "data" / "amazon0302.txt.gz"


graph = networkx.read_adjlist(DATA_FILE)


def bench_shortest_path():
    collections.deque(networkx.shortest_path_length(graph, "0"))


def bench_connected_components():
    networkx.number_connected_components(graph)


def bench_k_core():
    networkx.k_core(graph)


BENCHMARKS = {
    "shortest_path": bench_shortest_path,
    "connected_components": bench_connected_components,
    "k_core": bench_k_core,
}


def add_cmdline_args(cmd, args):
    cmd.append(args.benchmark)


def add_parser_args(parser):
    parser.add_argument("benchmark", choices=BENCHMARKS, help="Which benchmark to run.")


if __name__ == "__main__":
    runner = pyperf.Runner(add_cmdline_args=add_cmdline_args)
    runner.metadata["description"] = "NetworkX benchmark"
    add_parser_args(runner.argparser)
    args = runner.parse_args()
    benchmark = args.benchmark

    runner.bench_func(args.benchmark, BENCHMARKS[args.benchmark])



# Helpers for working with venvs.

import os
import os.path
import sys
import types

from . import _pip, _pythoninfo, _utils


class VenvCreationFailedError(Exception):
    def __init__(self, root, exitcode, already_existed):
        super().__init__(f"venv creation failed ({root})")
        self.root = root
        self.exitcode = exitcode
        self.already_existed = already_existed


class VenvPipInstallFailedError(Exception):
    def __init__(self, root, exitcode, msg=None):
        super().__init__(msg or f"failed to install pip in venv {root}")
        self.root = root
        self.exitcode = exitcode


class RequirementsInstallationFailedError(Exception):
    pass


def read_venv_config(root=None):
    """Return the config for the given venv, from its pyvenv.cfg file."""
    if not root:
        if sys.prefix == sys.base_prefix:
            raise Exception("current Python is not a venv")
        root = sys.prefix
    cfgfile = os.path.join(root, "pyvenv.cfg")
    with open(cfgfile, encoding="utf-8") as infile:
        text = infile.read()
    return parse_venv_config(text, root)


def parse_venv_config(lines, root=None):
    if isinstance(lines, str):
        lines = lines.splitlines()
    else:
        lines = (line.rstrip(os.linesep) for line in lines)

    cfg = types.SimpleNamespace(
        home=None,
        version=None,
        system_site_packages=None,
        prompt=None,
        executable=None,
        command=None,
    )
    fields = set(vars(cfg))
    for line in lines:
        # We do not validate the lines.
        name, sep, value = line.partition("=")
        if not sep:
            continue
        # We do not check for duplicate names.
        name = name.strip().lower()
        if name == "include-system-site-packages":
            name = "system_site_packages"
        if name not in fields:
            # XXX Preserve this anyway?
            continue
        value = value.lstrip()
        if name == "system_site_packages":
            value = value == "true"
        setattr(cfg, name, value)
    return cfg


def resolve_venv_python(root):
    python_exe = "python"
    if sys.executable.endswith(".exe"):
        python_exe += ".exe"
    if os.name == "nt":
        return os.path.join(root, "Scripts", python_exe)
    else:
        return os.path.join(root, "bin", python_exe)


def get_venv_root(name=None, venvsdir="venv", *, python=sys.executable):
    """Return the venv root to use for the given name (or given python)."""
    if not name:
        from .run import get_run_id

        runid = get_run_id(python)
        name = runid.name
    return os.path.abspath(
        os.path.join(venvsdir or ".", name),
    )


def venv_exists(root):
    venv_python = resolve_venv_python(root)
    return os.path.exists(venv_python)


def create_venv(
    root,
    python=sys.executable,
    *,
    env=None,
    downloaddir=None,
    withpip=True,
    cleanonfail=True,
):
    """Create a new venv at the given root, optionally installing pip."""
    already_existed = os.path.exists(root)
    if withpip:
        args = ["-m", "venv", root]
    else:
        args = ["-m", "venv", "--without-pip", root]
    ec, _, _ = _utils.run_python(*args, python=python, env=env)
    if ec != 0:
        if cleanonfail and not already_existed:
            _utils.safe_rmtree(root)
        raise VenvCreationFailedError(root, ec, already_existed)
    return resolve_venv_python(root)


class VirtualEnvironment:
    _env = None

    @classmethod
    def create(cls, root=None, python=sys.executable, **kwargs):
        if not python:
            python = sys.executable
        if isinstance(python, str):
            try:
                info = _pythoninfo.get_info(python)
            except FileNotFoundError:
                info = None
        else:
            info = python
        if not root:
            root = get_venv_root(python=info)

        print("Creating the virtual environment %s" % root)
        if venv_exists(root):
            raise Exception(f"virtual environment {root} already exists")

        try:
            venv_python = create_venv(root, info or python, **kwargs)
        except BaseException:
            _utils.safe_rmtree(root)
            raise  # re-raise
        if not info:
            info = _pythoninfo.get_info(python)
        self = cls(root, base=info)
        self._python = venv_python
        return self

    @classmethod
    def ensure(cls, root, python=None, **kwargs):
        if venv_exists(root):
            return cls(root)
        else:
            return cls.create(root, python, **kwargs)

    def __init__(self, root, *, base=None):
        assert os.path.exists(resolve_venv_python(root)), root
        self.root = root
        if base:
            self._base = base

    @property
    def python(self):
        try:
            return self._python
        except AttributeError:
            if not getattr(self, "_info", None):
                return resolve_venv_python(self.root)
            self._python = self.info.sys.executable
            return self._python

    @property
    def info(self):
        try:
            return self._info
        except AttributeError:
            try:
                python = self._python
            except AttributeError:
                python = resolve_venv_python(self.root)
            self._info = _pythoninfo.get_info(python)
            return self._info

    @property
    def base(self):
        try:
            return self._base
        except AttributeError:
            base_exe = self.info.sys._base_executable
            if not base_exe and base_exe != self.info.sys.executable:
                # XXX Use read_venv_config().
                raise NotImplementedError
                base_exe = ...
            self._base = _pythoninfo.get_info(base_exe)
            return self._base

    def ensure_pip(self, downloaddir=None, *, installer=True, upgrade=True):
        if not upgrade and _pip.is_pip_installed(self.python, env=self._env):
            return
        ec, _, _ = _pip.install_pip(
            self.python,
            info=self.info,
            downloaddir=downloaddir or self.root,
            env=self._env,
            upgrade=upgrade,
        )
        if ec != 0:
            raise VenvPipInstallFailedError(self.root, ec)
        elif not _pip.is_pip_installed(self.python, env=self._env):
            raise VenvPipInstallFailedError(self.root, 0, "pip doesn't work")

        if installer:
            # Upgrade installer dependencies (setuptools, ...)
            ec, _, _ = _pip.ensure_installer(
                self.python,
                env=self._env,
                upgrade=True,
            )
            if ec != 0:
                raise RequirementsInstallationFailedError("wheel")

    def upgrade_pip(self, *, installer=True):
        ec, _, _ = _pip.upgrade_pip(
            self.python,
            info=self.info,
            env=self._env,
            installer=installer,
        )
        if ec != 0:
            raise RequirementsInstallationFailedError("pip")

    def ensure_reqs(self, *reqs, upgrade=True):
        print("Installing requirements into the virtual environment %s" % self.root)
        ec, _, _ = _pip.install_requirements(
            *reqs,
            python=self.python,
            env=self._env,
            upgrade=upgrade,
        )
        if ec:
            raise RequirementsInstallationFailedError(reqs)



import re

from docutils import nodes
from docutils.parsers.rst import directives
from docutils.utils import unescape
from sphinx import addnodes
from sphinx.domains.changeset import VersionChange, versionlabels, versionlabel_classes
from sphinx.domains.python import PyFunction, PyMethod, PyModule
from sphinx.locale import _ as sphinx_gettext
from sphinx.util.docutils import SphinxDirective


ISSUE_URI = 'https://bugs.python.org/issue?@action=redirect&bpo=%s'
GH_ISSUE_URI = 'https://github.com/python/cpython/issues/%s'
# Used in conf.py and updated here by python/release-tools/run_release.py
SOURCE_URI = 'https://github.com/python/cpython/tree/3.13/%s'

# monkey-patch reST parser to disable alphabetic and roman enumerated lists
from docutils.parsers.rst.states import Body
Body.enum.converters['loweralpha'] = \
    Body.enum.converters['upperalpha'] = \
    Body.enum.converters['lowerroman'] = \
    Body.enum.converters['upperroman'] = lambda x: None

# monkey-patch the productionlist directive to allow hyphens in group names
# https://github.com/sphinx-doc/sphinx/issues/11854
from sphinx.domains import std

std.token_re = re.compile(r'`((~?[\w-]*:)?\w+)`')

# backport :no-index:
PyModule.option_spec['no-index'] = directives.flag


# Support for marking up and linking to bugs.python.org issues

def issue_role(typ, rawtext, text, lineno, inliner, options={}, content=[]):
    issue = unescape(text)
    # sanity check: there are no bpo issues within these two values
    if 47261 < int(issue) < 400000:
        msg = inliner.reporter.error(f'The BPO ID {text!r} seems too high -- '
                                     'use :gh:`...` for GitHub IDs', line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]
    text = 'bpo-' + issue
    refnode = nodes.reference(text, text, refuri=ISSUE_URI % issue)
    return [refnode], []


# Support for marking up and linking to GitHub issues

def gh_issue_role(typ, rawtext, text, lineno, inliner, options={}, content=[]):
    issue = unescape(text)
    # sanity check: all GitHub issues have ID >= 32426
    # even though some of them are also valid BPO IDs
    if int(issue) < 32426:
        msg = inliner.reporter.error(f'The GitHub ID {text!r} seems too low -- '
                                     'use :issue:`...` for BPO IDs', line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]
    text = 'gh-' + issue
    refnode = nodes.reference(text, text, refuri=GH_ISSUE_URI % issue)
    return [refnode], []


# Support for marking up implementation details

class ImplementationDetail(SphinxDirective):

    has_content = True
    final_argument_whitespace = True

    # This text is copied to templates/dummy.html
    label_text = sphinx_gettext('CPython implementation detail:')

    def run(self):
        self.assert_has_content()
        pnode = nodes.compound(classes=['impl-detail'])
        content = self.content
        add_text = nodes.strong(self.label_text, self.label_text)
        self.state.nested_parse(content, self.content_offset, pnode)
        content = nodes.inline(pnode[0].rawsource, translatable=True)
        content.source = pnode[0].source
        content.line = pnode[0].line
        content += pnode[0].children
        pnode[0].replace_self(nodes.paragraph(
            '', '', add_text, nodes.Text(' '), content, translatable=False))
        return [pnode]


# Support for documenting decorators

class PyDecoratorMixin(object):
    def handle_signature(self, sig, signode):
        ret = super(PyDecoratorMixin, self).handle_signature(sig, signode)
        signode.insert(0, addnodes.desc_addname('@', '@'))
        return ret

    def needs_arglist(self):
        return False


class PyDecoratorFunction(PyDecoratorMixin, PyFunction):
    def run(self):
        # a decorator function is a function after all
        self.name = 'py:function'
        return PyFunction.run(self)


# TODO: Use sphinx.domains.python.PyDecoratorMethod when possible
class PyDecoratorMethod(PyDecoratorMixin, PyMethod):
    def run(self):
        self.name = 'py:method'
        return PyMethod.run(self)


class PyCoroutineMixin(object):
    def handle_signature(self, sig, signode):
        ret = super(PyCoroutineMixin, self).handle_signature(sig, signode)
        signode.insert(0, addnodes.desc_annotation('coroutine ', 'coroutine '))
        return ret


class PyAwaitableMixin(object):
    def handle_signature(self, sig, signode):
        ret = super(PyAwaitableMixin, self).handle_signature(sig, signode)
        signode.insert(0, addnodes.desc_annotation('awaitable ', 'awaitable '))
        return ret


class PyCoroutineFunction(PyCoroutineMixin, PyFunction):
    def run(self):
        self.name = 'py:function'
        return PyFunction.run(self)


class PyCoroutineMethod(PyCoroutineMixin, PyMethod):
    def run(self):
        self.name = 'py:method'
        return PyMethod.run(self)


class PyAwaitableFunction(PyAwaitableMixin, PyFunction):
    def run(self):
        self.name = 'py:function'
        return PyFunction.run(self)


class PyAwaitableMethod(PyAwaitableMixin, PyMethod):
    def run(self):
        self.name = 'py:method'
        return PyMethod.run(self)


class PyAbstractMethod(PyMethod):

    def handle_signature(self, sig, signode):
        ret = super(PyAbstractMethod, self).handle_signature(sig, signode)
        signode.insert(0, addnodes.desc_annotation('abstractmethod ',
                                                   'abstractmethod '))
        return ret

    def run(self):
        self.name = 'py:method'
        return PyMethod.run(self)


# Support for documenting version of removal in deprecations

class DeprecatedRemoved(VersionChange):
    required_arguments = 2

    _deprecated_label = sphinx_gettext('Deprecated since version %s, will be removed in version %s')
    _removed_label = sphinx_gettext('Deprecated since version %s, removed in version %s')

    def run(self):
        # Replace the first two arguments (deprecated version and removed version)
        # with a single tuple of both versions.
        version_deprecated = self.arguments[0]
        version_removed = self.arguments.pop(1)
        self.arguments[0] = version_deprecated, version_removed

        # Set the label based on if we have reached the removal version
        current_version = tuple(map(int, self.config.version.split('.')))
        removed_version = tuple(map(int,  version_removed.split('.')))
        if current_version < removed_version:
            versionlabels[self.name] = self._deprecated_label
            versionlabel_classes[self.name] = 'deprecated'
        else:
            versionlabels[self.name] = self._removed_label
            versionlabel_classes[self.name] = 'removed'
        try:
            return super().run()
        finally:
            # reset versionlabels and versionlabel_classes
            versionlabels[self.name] = ''
            versionlabel_classes[self.name] = ''


def setup(app):
    app.add_role('issue', issue_role)
    app.add_role('gh', gh_issue_role)
    app.add_directive('impl-detail', ImplementationDetail)
    app.add_directive('deprecated-removed', DeprecatedRemoved)
    app.add_directive_to_domain('py', 'decorator', PyDecoratorFunction)
    app.add_directive_to_domain('py', 'decoratormethod', PyDecoratorMethod)
    app.add_directive_to_domain('py', 'coroutinefunction', PyCoroutineFunction)
    app.add_directive_to_domain('py', 'coroutinemethod', PyCoroutineMethod)
    app.add_directive_to_domain('py', 'awaitablefunction', PyAwaitableFunction)
    app.add_directive_to_domain('py', 'awaitablemethod', PyAwaitableMethod)
    app.add_directive_to_domain('py', 'abstractmethod', PyAbstractMethod)
    return {'version': '1.0', 'parallel_read_safe': True}



"""
Benchmark constant folding over a module of integer addition of constants using
xDSL's pattern rewriting machinery. The workload module is generated by the
``get_constant_folding_module()`` function, which is then transformed by the
``xdsl.transforms.test_constant_folding.TestConstantFoldingPass`` xDSL
transformation pass.

xDSL is a Python-native compiler framework built around SSA-based intermediate
representations. It re-implements many of MLIR's data structures and methods in
Python. This benchmark exercises the simple pattern rewriting transformation of
constant folding. This is a fair proxy for pattern rewriting transformations in
general, which are a major component of MLIR-like compilers in lowering passes.

Pattern rewriting in both xDSL and MLIR is a pointer-chasing, unstructured
workload, which makes it hard to optimise ahead-of-time. This diminishes the
traditional performance advantage of ahead-of-time compiled languages such as
C++ over dynamic languages such as Python -- making it an interesting benchmark.
More information about the design and impact of this benchmark can be found in
the Master's thesis ``Performance and Dynamism in User-extensible Compiler
Infrastructures'', which is `available on GitHub
<https://github.com/EdmundGoodman/masters-project>`_.
"""

import random

import pyperf

from xdsl.context import Context
from xdsl.dialects.arith import AddiOp, Arith, ConstantOp
from xdsl.dialects.builtin import (
    Builtin,
    IntegerAttr,
    ModuleOp,
    i32,
)
from xdsl.dialects.test import TestOp
from xdsl.ir import Operation
from xdsl.transforms.test_constant_folding import TestConstantFoldingPass


RANDOM_SEED = 0


def get_constant_folding_module(size: int = 100) -> ModuleOp:
    """Generate an integer addition constant folding workload of a given size.

    The output of running the command
    `print(WorkloadBuilder().constant_folding_module(size=5))` is shown
    below:

    ```mlir
    "builtin.module"() ({
        %0 = "arith.constant"() {"value" = 865 : i32} : () -> i32
        %1 = "arith.constant"() {"value" = 395 : i32} : () -> i32
        %2 = "arith.addi"(%1, %0) : (i32, i32) -> i32
        %3 = "arith.constant"() {"value" = 777 : i32} : () -> i32
        %4 = "arith.addi"(%3, %2) : (i32, i32) -> i32
        %5 = "arith.constant"() {"value" = 912 : i32} : () -> i32
        "test.op"(%4) : (i32) -> ()
    }) : () -> ()
    ```
    """
    assert size >= 0
    random.seed(RANDOM_SEED)
    ops: list[Operation] = []
    ops.append(ConstantOp(IntegerAttr(random.randint(1, 1000), i32)))
    for i in range(1, size + 1):
        if i % 2 == 0:
            ops.append(AddiOp(ops[i - 1], ops[i - 2]))
        else:
            ops.append(ConstantOp(IntegerAttr(random.randint(1, 1000), i32)))
    ops.append(TestOp([ops[(size // 2) * 2]]))
    return ModuleOp(ops)


def bench_xdsl_constant_folding(loops: int, size: int) -> float:
    """Benchmark constant folding integer additions with xDSL."""
    ctx = Context(allow_unregistered=True)
    ctx.load_dialect(Arith)
    ctx.load_dialect(Builtin)
    constant_folding_pass = TestConstantFoldingPass()
    constant_folding_workload = get_constant_folding_module(size)

    range_it = range(loops)
    t0 = pyperf.perf_counter()
    for _ in range_it:
        constant_folding_workload_mutable = constant_folding_workload.clone()
        constant_folding_pass.apply(ctx, constant_folding_workload_mutable)
    return pyperf.perf_counter() - t0


def add_cmdline_args(cmd, args):
    """Add command line arguments to a pyperf runner"""
    cmd.extend(("--size", str(args.size)))


if __name__ == "__main__":
    runner = pyperf.Runner(add_cmdline_args=add_cmdline_args)
    runner.metadata["description"] = "Benchmark xDSL constant folding integer addition"
    runner.argparser.add_argument(
        "--size",
        type=int,
        default=1000,
        help="Number of integer additions (default: 1000)",
    )
    args = runner.parse_args()
    runner.bench_time_func("xdsl_constant_fold", bench_xdsl_constant_folding, args.size)



# This module should be replaced with the equivalent functionality
# in the PyPI "packaging" package (once it's added there).

__all__ = [
    "load_pyproject_toml",
    "parse_classifier",
    "parse_entry_point",
    "parse_person",
    "parse_pyproject_toml",
]


import os.path
import re
import urllib.parse

import packaging.requirements
import packaging.specifiers
import packaging.utils
import packaging.version

try:
    import tomllib  # type: ignore[import-not-found] # tomllib doesn't exist on 3.7-3.10
except ImportError:
    import tomli as tomllib

from ._utils import check_name

NAME_RE = re.compile("^([A-Z0-9]|[A-Z0-9][A-Z0-9._-]*[A-Z0-9])$", re.IGNORECASE)


def parse_person(text):
    # XXX
    return text


def parse_classifier(text):
    # XXX Use https://pypi.org/project/packaging-classifiers.
    return text


def parse_entry_point(text):
    # See:
    #  * https://packaging.python.org/specifications/entry-points/#data-model
    #  * https://www.python.org/dev/peps/pep-0517/#source-trees
    module, sep, qualname = text.partition(":")
    if all(p.isidentifier() for p in module.split(".")):
        if not sep or all(p.isidentifier() for p in qualname.split(".")):
            return module, qualname

    raise ValueError(f"invalid entry point {text!r}")


def parse_pyproject_toml(
    text,
    rootdir,
    name=None,
    *,
    tools=None,
    requirefiles=True,
):
    data = tomllib.loads(text)
    unused = list(data)

    for section, normalize in SECTIONS.items():
        try:
            secdata = data[section]
        except KeyError:
            data[section] = None
        else:
            data[section] = normalize(
                secdata,
                name=name,
                tools=tools,
                rootdir=rootdir,
                requirefiles=requirefiles,
            )
            unused.remove(section)

    if unused:
        raise ValueError(f"unsupported sections ({', '.join(sorted(unused))})")

    return data


def load_pyproject_toml(filename, *, name=None, tools=None, requirefiles=True):
    if os.path.isdir(filename):
        rootdir = filename
        filename = os.path.join(rootdir, "pyproject.toml")
    else:
        rootdir = os.path.dirname(filename)

    with open(filename, encoding="utf-8") as infile:
        text = infile.read()
    data = parse_pyproject_toml(
        text,
        rootdir,
        name,
        tools=tools,
        requirefiles=requirefiles,
    )
    return data, filename


#######################################
# internal implementation


def _check_relfile(relname, rootdir, kind):
    if os.path.isabs(relname):
        raise ValueError(f"{relname!r} is absolute, expected relative")
    actual = os.path.join(rootdir, relname)
    if kind == "dir":
        if not os.path.isdir(actual):
            raise ValueError(f"directory {actual!r} does not exist")
    elif kind == "file":
        if not os.path.isfile(actual):
            raise ValueError(f"file {actual!r} does not exist")
    elif kind == "any":
        if not os.path.exists(actual):
            raise ValueError(f"{actual!r} does not exist")
    elif kind:
        raise NotImplementedError(kind)


def _check_file_or_text(table, rootdir, requirefiles, extra=None):
    unsupported = set(table) - set(["file", "text"]) - set(extra or ())
    if unsupported:
        raise ValueError(f"unsupported license data {table!r}")

    if "file" in table:
        if "text" in table:
            raise ValueError('"file" and "text" are mutually exclusive')
        kind = "file" if requirefiles else None
        _check_relfile(table["file"], rootdir, kind)


def _normalize_project(data, rootdir, name, requirefiles, **_ignored):
    # See PEP 621.
    unused = set(data)

    ##########
    # First handle the required fields.

    name = data.get("name", name)
    if name:
        if not NAME_RE.match(name):
            raise ValueError(f"invalid name {name!r}")
        name = packaging.utils.canonicalize_name(name)
        data["name"] = name
        if "name" in unused:
            unused.remove("name")
    else:
        if "name" not in data.get("dynamic", []):
            raise ValueError('missing required "name" field')

    try:
        version = data["version"]
    except KeyError:
        if "version" not in data.get("dynamic", []):
            raise ValueError('missing required "version" field')
    else:
        # We keep the full version string rather than
        # the canonicalized form.  However, we still validate and
        # (effectively) normalize it.
        version = packaging.version.parse(version)
        data["version"] = str(version)
        unused.remove("version")

    ##########
    # Now we handle the optional fields.

    # We leave "description" as-is.

    key = "readme"
    if key in data:
        readme = data[key]
        if isinstance(readme, str):
            readme = data[key] = {"file": readme}
        # XXX Check the suffix.
        # XXX Handle 'content-type'.
        # XXX Handle "charset" parameter.
        _check_file_or_text(
            data[key], rootdir, requirefiles, ["content-type", "charset"]
        )
        unused.remove(key)

    key = "requires-python"
    if key in data:
        # We keep it as a string.
        data[key] = str(packaging.specifiers.SpecifierSet(data[key]))
        unused.remove(key)

    key = "license"
    if key in data:
        _check_file_or_text(data[key], rootdir, requirefiles)
        unused.remove(key)

    key = "keywords"
    if key in data:
        for keyword in data[key]:
            # XXX Is this the right check?
            check_name(name, loose=True)
        unused.remove(key)

    key = "authors"
    if key in data:
        for person in data[key]:
            # We only make sure it is valid.
            parse_person(person)
        unused.remove(key)

    key = "maintainers"
    if key in data:
        for person in data[key]:
            # We only make sure it is valid.
            parse_person(person)
        unused.remove(key)

    key = "classifiers"
    if key in data:
        for classifier in data[key]:
            # We only make sure it is valid.
            parse_classifier(classifier)
        unused.remove(key)

    key = "dependencies"
    if key in data:
        for dep in data[key]:
            # We only make sure it is valid.
            packaging.requirements.Requirement(dep)
        unused.remove(key)

    key = "optional-dependencies"
    if key in data:
        # XXX
        unused.remove(key)

    key = "urls"
    if key in data:
        for name, url in data[key].items():
            # XXX Is there a stricter check?
            check_name(name, loose=True)
            # We only make sure it is valid.
            urllib.parse.urlparse(url)
        unused.remove(key)

    key = "scripts"
    if key in data:
        for name, value in data[key].items():
            # XXX Is there a stricter check?
            check_name(name, loose=True)
            # We only make sure it is valid.
            parse_entry_point(value)
        unused.remove(key)

    key = "gui-scripts"
    if key in data:
        for _, value in data[key].items():
            # XXX Is there a stricter check?
            check_name(name, loose=True)
            # We only make sure it is valid.
            parse_entry_point(value)
        unused.remove(key)

    key = "entry-points"
    if key in data:
        for groupname, group in data[key].items():
            # XXX Is there a stricter check?
            check_name(groupname, loose=True)
            for epname, value in group.items():
                # XXX Is there a stricter check?
                check_name(epname, loose=True)
                # We only make sure it is valid.
                parse_entry_point(value)
        unused.remove(key)

    key = "dynamic"
    if key in data:
        for field in data[key]:
            check_name(field, loose=True)
            # XXX Fail it isn't one of the supported fields.
        unused.remove(key)

    return data


def _normalize_build_system(data, rootdir, requirefiles, **_ignored):
    # See PEP 518 and 517.
    unused = set(data)

    key = "requires"
    if key in data:
        reqs = data[key]
        for i, raw in enumerate(reqs):
            # We only make sure it is valid.
            packaging.requirements.Requirement(raw)
        unused.remove(key)
    else:
        raise ValueError('missing "requires" field')

    key = "build-backend"
    if key in data:
        # We only make sure it is valid.
        parse_entry_point(data[key])
        unused.remove(key)

    key = "backend-path"
    if key in data:
        if "build-backend" not in data:
            raise ValueError('missing "build-backend" field')
        kind = "dir" if requirefiles else None
        for dirname in data[key]:
            _check_relfile(dirname, rootdir, kind=kind)
        unused.remove(key)

    if unused:
        raise ValueError(f"unsupported keys ({', '.join(sorted(unused))})")

    return data


def _normalize_tool(data, tools, rootdir, **_ignored):
    # See PEP 518.
    tools = tools or {}
    for name, tooldata in list(data.items()):
        if name in tools:
            normalize = tools[name]
            data[name] = normalize(name, tooldata, rootdir=rootdir)
            if data[name] is None:
                del data[name]
    return data


SECTIONS = {
    "project": _normalize_project,
    "build-system": _normalize_build_system,
    "tool": _normalize_tool,
}



"""
Benchmark argparse programs with:
1) multiple subparsers, each with their own subcommands, and then parse a series of command-line arguments.
2) a large number of optional arguments, and then parse a series of command-line arguments.

Author: Savannah Ostrowski
"""

import argparse
import pyperf


def generate_arguments(i: int) -> list:
    arguments = ["input.txt", "output.txt"]
    for i in range(i):
        arguments.extend([f"--option{i}", f"value{i}"])
    return arguments


def bm_many_optionals() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(description="A version control system CLI")

    parser.add_argument("--version", action="version", version="1.0")

    subparsers = parser.add_subparsers(dest="command", required=True)

    add_parser = subparsers.add_parser("add", help="Add a file to the repository")
    add_parser.add_argument("files", nargs="+", help="List of files to add to staging")

    commit_parser = subparsers.add_parser(
        "commit", help="Commit changes to the repository"
    )
    commit_parser.add_argument("-m", "--message", required=True, help="Commit message")

    commit_group = commit_parser.add_mutually_exclusive_group(required=False)
    commit_group.add_argument(
        "--amend", action="store_true", help="Amend the last commit"
    )
    commit_group.add_argument(
        "--no-edit", action="store_true", help="Reuse the last commit message"
    )

    push_parser = subparsers.add_parser(
        "push", help="Push changes to remote repository"
    )

    network_group = push_parser.add_argument_group("Network options")
    network_group.add_argument("--dryrun", action="store_true", help="Simulate changes")
    network_group.add_argument(
        "--timeout", type=int, default=30, help="Timeout in seconds"
    )

    auth_group = push_parser.add_argument_group("Authentication options")
    auth_group.add_argument(
        "--username", required=True, help="Username for authentication"
    )
    auth_group.add_argument(
        "--password", required=True, help="Password for authentication"
    )

    global_group = parser.add_mutually_exclusive_group()
    global_group.add_argument("--verbose", action="store_true", help="Verbose output")
    global_group.add_argument("--quiet", action="store_true", help="Quiet output")

    argument_lists = [
        ["--verbose", "add", "file1.txt", "file2.txt"],
        ["add", "file1.txt", "file2.txt"],
        ["commit", "-m", "Initial commit"],
        ["commit", "-m", "Add new feature", "--amend"],
        [
            "push",
            "--dryrun",
            "--timeout",
            "60",
            "--username",
            "user",
            "--password",
            "pass",
        ],
    ]

    for arguments in argument_lists:
        parser.parse_args(arguments)


def bm_subparsers() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()

    parser.add_argument("input_file", type=str, help="The input file")
    parser.add_argument("output_file", type=str, help="The output file")

    for i in range(1000):
        parser.add_argument(f"--option{i}", type=str, help=f"Optional argument {i}")

    argument_lists = [
        generate_arguments(500),
        generate_arguments(1000),
    ]

    for args in argument_lists:
        parser.parse_args(args)


BENCHMARKS = {
    "many_optionals": bm_many_optionals,
    "subparsers": bm_subparsers,
}


def add_cmdline_args(cmd, args):
    cmd.append(args.benchmark)


def add_parser_args(parser):
    parser.add_argument("benchmark", choices=BENCHMARKS, help="Which benchmark to run.")


if __name__ == "__main__":
    runner = pyperf.Runner(add_cmdline_args=add_cmdline_args)
    runner.metadata["description"] = "Argparse benchmark"
    add_parser_args(runner.argparser)
    args = runner.parse_args()
    benchmark = args.benchmark

    runner.bench_func(args.benchmark, BENCHMARKS[args.benchmark])



__all__ = [
    "BenchmarksManifest",
    "load_manifest",
    "parse_manifest",
]


import os.path

from . import DATA_DIR, __version__, _benchmark, _utils

DEFAULTS_DIR = os.path.join(DATA_DIR, "benchmarks")
DEFAULT_MANIFEST = os.path.join(DEFAULTS_DIR, "MANIFEST")

BENCH_COLUMNS = ("name", "metafile")
BENCH_HEADER = "\t".join(BENCH_COLUMNS)


def load_manifest(filename, *, resolve=None):
    if not filename:
        filename = DEFAULT_MANIFEST
    filename = _utils.resolve_file(filename)
    sections = _parse_manifest_file(filename)
    return BenchmarksManifest._from_sections(sections, resolve, filename)


def parse_manifest(lines, *, resolve=None, filename=None):
    if isinstance(lines, str):
        lines = lines.splitlines()
    else:
        if not filename:
            # Try getting the filename from a file.
            filename = getattr(lines, "name", None)
    sections = _parse_manifest(lines, filename)
    return BenchmarksManifest._from_sections(sections, resolve, filename)


def resolve_default_benchmark(bench):
    if isinstance(bench, _benchmark.Benchmark):
        spec = bench.spec
    else:
        spec = bench
        bench = _benchmark.Benchmark(spec, "<bogus>")
        bench.metafile = None

    if not spec.version:
        spec = spec._replace(version=__version__)
    if not spec.origin:
        spec = spec._replace(origin="<default>")
    bench.spec = spec

    if not bench.metafile:
        metafile = os.path.join(DEFAULTS_DIR, f"bm_{bench.name}", "pyproject.toml")
        bench.metafile = metafile
    return bench


class BenchmarksManifest:
    @classmethod
    def _from_sections(cls, sections, resolve=None, filename=None):
        self = cls(filename=filename)
        self._add_sections(sections, resolve)
        return self

    def __init__(self, benchmarks=None, groups=None, filename=None):
        self._raw_benchmarks = []
        # XXX Support disabling all groups (except all and default)?
        self._raw_groups = {}
        self._raw_filename = filename
        self._byname = {}
        self._groups = None
        self._tags = None

        if benchmarks:
            self._add_benchmarks(benchmarks)
        if groups:
            self._add_groups(groups)

    def __repr__(self):
        args = (
            f"{n}={getattr(self, '_raw_' + n)}"
            for n in ("benchmarks", "groups", "filename")
        )
        return f"{type(self).__name__}({', '.join(args)})"

    @property
    def benchmarks(self):
        return list(self._byname.values())

    @property
    def groups(self):
        names = self._custom_groups()
        return names | {"all", "default"}

    @property
    def tags(self):
        return set(self._get_tags())

    @property
    def filename(self):
        return self._raw_filename

    def _add_sections(self, sections, resolve):
        seen_by_file = {}
        for filename, section, data in sections:
            try:
                seen = seen_by_file[filename]
            except KeyError:
                seen = seen_by_file[filename] = set()
            self._add_section_for_file(filename, section, data, resolve, seen)

    def _add_section_for_file(self, filename, section, data, resolve, seen):
        if resolve is None and filename == DEFAULT_MANIFEST:
            resolve = resolve_default_benchmark

        if section == "group":
            name, entries = data
            self._add_group(name, entries)
        else:
            # All sections with an identifier have already been handled.
            if section in seen:
                # For now each section_key can only show up once.
                raise NotImplementedError((section, data))
            seen.add(section)

            if section == "includes":
                pass
            elif section == "benchmarks":
                entries = ((s, m, filename) for s, m in data)
                self._add_benchmarks(entries, resolve)
            elif section == "groups":
                for name in data:
                    self._add_group(name, None)
            else:
                raise NotImplementedError((section, data))

    def _add_benchmarks(self, entries, resolve):
        for spec, metafile, filename in entries:
            # XXX Ignore duplicates?
            self._add_benchmark(spec, metafile, resolve, filename)

    def _add_benchmark(self, spec, metafile, resolve, filename):
        if spec.name in self._raw_groups:
            raise ValueError(
                f"a group and a benchmark have the same name ({spec.name})"
            )
        if spec.name == "all":
            raise ValueError(
                'a benchmark named "all" is not allowed ("all" is reserved for selecting the full set of declared benchmarks)'
            )
        if metafile:
            if filename:
                localdir = os.path.dirname(filename)
                metafile = os.path.join(localdir, metafile)
            bench = _benchmark.Benchmark(spec, metafile)
        else:
            metafile = None
            bench = spec
        self._raw_benchmarks.append((spec, metafile, filename))
        if resolve is not None:
            bench = resolve(bench)
        if bench.name in self._byname:
            raise ValueError(f"a benchmark named {bench.name} was already declared")
        self._byname[bench.name] = bench
        self._groups = None  # Force re-resolution.
        self._tags = None  # Force re-resolution.

    def _add_group(self, name, entries):
        if name in self._byname:
            raise ValueError(f"a group and a benchmark have the same name ({name})")
        if name == "all":
            raise ValueError(
                'a group named "all" is not allowed ("all" is reserved for selecting the full set of declared benchmarks)'
            )
        if entries is None:
            if name in self._raw_groups:
                return
            self._raw_groups[name] = None
        elif name in self._raw_groups and self._raw_groups[name] is not None:
            raise ValueError(f"a group named {name} was already defined")
        else:
            self._raw_groups[name] = list(entries) if entries else []
        self._groups = None  # Force re-resolution.

    def _custom_groups(self):
        return set(self._raw_groups) - {"all", "default"}

    def _get_tags(self):
        if self._tags is None:
            self._tags = _get_tags(self._byname.values())
            self._tags.pop("all", None)  # It is manifest-specific.
            self._tags.pop("default", None)  # It is manifest-specific.
        return self._tags

    def _resolve_groups(self):
        if self._groups is not None:
            return self._groups

        raw = {}
        for name, entries in self._raw_groups.items():
            if entries and entries[0][0] == "-":
                entries = list(entries)
                entries.insert(0, ("+", "<all>"))
            raw[name] = entries
        self._groups = _resolve_groups(raw, self._byname)
        return self._groups

    def resolve_group(self, name, *, fail=True):
        if name == "all":
            benchmarks = self._byname.values()
        elif name == "default":
            if "default" not in self._raw_groups:
                benchmarks = self._byname.values()
            else:
                groups = self._resolve_groups()
                benchmarks = groups.get(name)
        elif not self._custom_groups():
            benchmarks = self._get_tags().get(name)
            if benchmarks is None and fail:
                raise KeyError(name)
        else:
            groups = self._resolve_groups()
            benchmarks = groups.get(name)
            if not benchmarks:
                if name not in self._raw_groups:
                    benchmarks = self._get_tags().get(name, ())
                elif fail:
                    raise KeyError(name)
        yield from benchmarks or ()

    def show(self, *, raw=True, resolved=True):
        yield self.filename
        yield "groups:"
        if raw:
            yield f" {self._raw_groups}"
        if resolved:
            yield f" {self.groups}"
        yield "default:"
        if resolved:
            for i, bench in enumerate(self.resolve_group("default")):
                yield f" {i:>2} {bench}"
        if raw:
            yield "benchmarks (raw):"
            for i, bench in enumerate(self._raw_benchmarks):
                yield f" {i:>2} {bench}"
        if resolved:
            yield "benchmarks:"
            for i, bench in enumerate(self.benchmarks):
                yield f" {i:>2} {bench}"


#######################################
# internal implementation


def _iter_sections(lines):
    lines = (line.split("#")[0].strip() for line in lines)

    name = None
    section = None
    for line in lines:
        if not line:
            continue
        if line.startswith("[") and line.endswith("]"):
            if name:
                yield name, section
            name = line[1:-1].strip()
            section = []
        else:
            if not name:
                raise ValueError(f"expected new section, got {line!r}")
            section.append(line)
    if name:
        yield name, section
    else:
        raise ValueError("invalid manifest file, no sections found")


def _parse_manifest_file(filename):
    with open(filename, encoding="utf-8") as infile:
        yield from _parse_manifest(infile, filename)


def _parse_manifest(lines, filename):
    relroot = os.path.dirname(filename)
    for section, seclines in _iter_sections(lines):
        if section == "includes":
            yield filename, section, list(seclines)
            for line in seclines:
                if line == "<default>":
                    line = DEFAULT_MANIFEST
                else:
                    line = _utils.resolve_file(line, relroot)
                yield from _parse_manifest_file(line)
        elif section == "benchmarks":
            yield filename, section, list(_parse_benchmarks_section(seclines))
        elif section == "groups":
            yield filename, section, list(_parse_groups_section(seclines))
        elif section.startswith("group "):
            section, _, group = section.partition(" ")
            entries = list(_parse_group_section(seclines))
            yield filename, section, (group, entries)
        else:
            raise ValueError(f"unsupported section {section!r}")


def _parse_benchmarks_section(lines):
    if not lines:
        lines = ["<empty>"]
    lines = iter(lines)
    if next(lines) != BENCH_HEADER:
        raise ValueError("invalid manifest file, expected benchmarks table header")

    version = origin = None
    for line in lines:
        try:
            name, metafile = (
                None if field == "-" else field for field in line.split("\t")
            )
        except ValueError:
            raise ValueError(f"bad benchmark line {line!r}")
        spec = _benchmark.BenchmarkSpec(name or None, version, origin)
        metafile = _parse_metafile(metafile, name)
        yield spec, metafile


def _parse_metafile(metafile, name):
    if not metafile:
        return None
    elif metafile.startswith("<") and metafile.endswith(">"):
        directive, _, extra = metafile[1:-1].partition(":")
        if directive == "local":
            if extra:
                rootdir = f"bm_{extra}"
                basename = f"bm_{name}.toml"
            else:
                rootdir = f"bm_{name}"
                basename = "pyproject.toml"
            # A relative path will be resolved against the manifset file.
            return os.path.join(rootdir, basename)
        else:
            raise ValueError(f"unsupported metafile directive {metafile!r}")
    else:
        return os.path.abspath(metafile)


def _parse_groups_section(lines):
    for name in lines:
        _utils.check_name(name)
        yield name


def _parse_group_section(lines):
    for line in lines:
        if line.startswith("-"):
            # Exclude a benchmark or group.
            op = "-"
            name = line[1:]
        elif line.startswith("+"):
            op = "+"
            name = line[1:]
        else:
            op = "+"
            name = line
        _benchmark.check_name(name)
        yield op, name


def _get_tags(benchmarks):
    # Fill in groups from benchmark tags.
    tags = {}
    for bench in benchmarks:
        for tag in getattr(bench, "tags", ()):
            if tag in tags:
                tags[tag].append(bench)
            else:
                tags[tag] = [bench]
    return tags


def _resolve_groups(rawgroups, byname):
    benchmarks = set(byname.values())
    tags = None
    groups = {
        "all": list(benchmarks),
    }
    unresolved = {}
    for groupname, entries in rawgroups.items():
        if groupname == "all":
            continue
        if not entries:
            if groupname == "default":
                groups[groupname] = list(benchmarks)
            else:
                if tags is None:
                    tags = _get_tags(benchmarks)
                groups[groupname] = tags.get(groupname, ())
            continue
        assert entries[0][0] == "+", (groupname, entries)
        unresolved[groupname] = names = set()
        for op, name in entries:
            if op == "+":
                if name == "<all>":
                    names.update(byname)
                elif name in byname or name in rawgroups:
                    names.add(name)
            elif op == "-":
                if name == "<all>":
                    raise NotImplementedError((groupname, op, name))
                elif name in byname or name in rawgroups:
                    if name in names:
                        names.remove(name)
            else:
                raise NotImplementedError((groupname, op, name))
    while unresolved:
        for groupname, names in list(unresolved.items()):
            benchmarks = set()

            q = list(names)
            while q:
                name = q.pop()

                if name in byname:
                    benchmarks.add(byname[name])
                elif name in groups:
                    benchmarks.update(groups[name])
                elif name == groupname:
                    pass
                else:  # name in unresolved
                    q.extend(unresolved[name])

            groups[groupname] = benchmarks
            del unresolved[groupname]
    return groups



from urllib.request import urlopen
import json
import toml

BASE_URL = "https://api.github.com/repos/python/cpython/pulls?per_page=1000&state=all"

def main():
    all_issues = []
    for page in range(1, 11):
        with urlopen(f"{BASE_URL}&page={page}") as response:
            issues = json.loads(response.read())
            if not issues:
                break
            all_issues.extend(issues)
            print(f"Page: {page} Total Issues: {len(all_issues)}")
    with open("issues.toml", "w") as f:
        f.write(toml.dumps({"data": all_issues}))

if __name__ == "__main__":
    main()



import types
import unittest

from pyperformance import _python


class GetIDTests(unittest.TestCase):
    def _dummy_info(self):
        info = types.SimpleNamespace(
            sys=types.SimpleNamespace(
                executable="/a/b/c/bin/spam-python",
                version="3.8.10 (default, May  5 2021, 03:01:07) \n[GCC 7.5.0]",
                version_info=(3, 8, 10, "final", 0),
                api_version=1013,
                implementation=types.SimpleNamespace(
                    name="cpython",
                    version=(3, 8, 10, "final", 0),
                ),
            ),
            pyc_magic_number=b"U\r\r\n",
        )
        base_id = "b14d92fd0e6f"
        return info, base_id

    def test_no_prefix(self):
        info, expected = self._dummy_info()

        pyid = _python.get_id(info)

        self.assertEqual(pyid, expected)

    def test_default_prefix(self):
        info, expected = self._dummy_info()
        expected = f"cpython3.8-{expected}"

        pyid = _python.get_id(info, prefix=True)

        self.assertEqual(pyid, expected)

    def test_given_prefix(self):
        info, expected = self._dummy_info()
        expected = f"spam-{expected}"

        pyid = _python.get_id(info, prefix="spam-")

        self.assertEqual(pyid, expected)



"""
Benchmark a BPE tokeniser.

Author: Shantanu Jain

Based on code from tiktoken.
https://github.com/openai/tiktoken
"""
from __future__ import annotations

import pyperf


import collections
import re
from pathlib import Path


class SimpleBytePairEncoding:
    def __init__(self, *, pat_str: str, mergeable_ranks: dict[bytes, int]) -> None:
        self.pat_str = pat_str
        self.mergeable_ranks = mergeable_ranks

        self._decoder = {token: token_bytes for token_bytes, token in mergeable_ranks.items()}
        self._pat = re.compile(pat_str)

    def encode(self, text: str) -> list[int]:
        # Use the regex to split the text into (approximately) words
        words = self._pat.findall(text)
        tokens = []
        for word in words:
            # Turn each word into tokens, using the byte pair encoding algorithm
            word_bytes = word.encode("utf-8")
            word_tokens = bpe_encode(self.mergeable_ranks, word_bytes)
            tokens.extend(word_tokens)
        return tokens

    def decode_bytes(self, tokens: list[int]) -> bytes:
        return b"".join(self._decoder[token] for token in tokens)

    def decode(self, tokens: list[int]) -> str:
        return self.decode_bytes(tokens).decode("utf-8", errors="replace")

    @staticmethod
    def train(training_data: str, vocab_size: int, pat_str: str):
        mergeable_ranks = bpe_train(data=training_data, vocab_size=vocab_size, pat_str=pat_str)
        return SimpleBytePairEncoding(pat_str=pat_str, mergeable_ranks=mergeable_ranks)


def bpe_encode(mergeable_ranks: dict[bytes, int], input: bytes) -> list[int]:
    # A simple, uncached, quadratic BPE
    parts = [bytes([b]) for b in input]
    while True:
        # Iterate over all pairs and find the pair we want to merge the most
        min_idx = None
        min_rank = None
        for i, pair in enumerate(zip(parts[:-1], parts[1:])):
            rank = mergeable_ranks.get(pair[0] + pair[1])
            if rank is not None and (min_rank is None or rank < min_rank):
                min_idx = i
                min_rank = rank

        # If there were no pairs we could merge, we're done!
        if min_rank is None:
            break
        assert min_idx is not None

        # Otherwise, merge that pair and leave the rest unchanged. Then repeat.
        parts = parts[:min_idx] + [parts[min_idx] + parts[min_idx + 1]] + parts[min_idx + 2 :]

    tokens = [mergeable_ranks[part] for part in parts]
    return tokens


def bpe_train(data: str, vocab_size: int, pat_str: str) -> dict[bytes, int]:
    # First, add tokens for each individual byte value
    if vocab_size < 2**8:
        raise ValueError("vocab_size must be at least 256, so we can encode all bytes")
    ranks = {}
    for i in range(2**8):
        ranks[bytes([i])] = i

    # Splinter up our data into lists of bytes
    # data = "Hello world"
    # words = [
    #     [b'H', b'e', b'l', b'l', b'o'],
    #     [b' ', b'w', b'o', b'r', b'l', b'd']
    # ]
    words: list[list[bytes]] = [
        [bytes([b]) for b in word.encode("utf-8")] for word in re.findall(pat_str, data)
    ]

    # Now, use our data to figure out which merges we should make
    while len(ranks) < vocab_size:
        # Find the most common pair. This will become our next token
        stats = collections.Counter()
        for piece in words:
            for pair in zip(piece[:-1], piece[1:]):
                stats[pair] += 1

        most_common_pair = max(stats, key=lambda x: stats[x])
        token_bytes = most_common_pair[0] + most_common_pair[1]
        token = len(ranks)
        # Add the new token!
        ranks[token_bytes] = token

        # Now merge that most common pair in all the words. That is, update our training data
        # to reflect our decision to make that pair into a new token.
        new_words = []
        for word in words:
            new_word = []
            i = 0
            while i < len(word) - 1:
                if (word[i], word[i + 1]) == most_common_pair:
                    # We found our pair! Merge it
                    new_word.append(token_bytes)
                    i += 2
                else:
                    new_word.append(word[i])
                    i += 1
            if i == len(word) - 1:
                new_word.append(word[i])
            new_words.append(new_word)
        words = new_words

    return ranks


def train(data: str):
    pattern = (
        r"""'s|'t|'re|'ve|'m|'ll|'d| ?[a-zA-Z]+| ?\d+| ?[^\sa-zA-Z\d]+|\s+(?!\S)|\s+"""
    )
    enc = SimpleBytePairEncoding.train(data, vocab_size=1024, pat_str=pattern)

    tokens = enc.encode("hello world")
    assert enc.decode(tokens) == "hello world"

    enc.encode(data)


def bench_bpe_tokeniser(loops: int) -> float:
    DATA = Path(__file__).parent / "data" / "frankenstein_intro.txt"
    with open(DATA, "r", encoding="utf8") as f:
        data = f.read()

    range_it = range(loops)

    t0 = pyperf.perf_counter()
    for _ in range_it:
        train(data)
    return pyperf.perf_counter() - t0


if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata["description"] = "Benchmark a BPE tokeniser"
    runner.bench_time_func("bpe_tokeniser", bench_bpe_tokeniser)



"""
MathWorld: "Hundred-Dollar, Hundred-Digit Challenge Problems", Challenge #3.
http://mathworld.wolfram.com/Hundred-DollarHundred-DigitChallengeProblems.html

The Computer Language Benchmarks Game
http://benchmarksgame.alioth.debian.org/u64q/spectralnorm-description.html#spectralnorm

Contributed by Sebastien Loisel
Fixed by Isaac Gouy
Sped up by Josh Goldfoot
Dirtily sped up by Simon Descarpentries
Concurrency by Jason Stitt
"""

import pyperf


DEFAULT_N = 130


def eval_A(i, j):
    return 1.0 / ((i + j) * (i + j + 1) // 2 + i + 1)


def eval_times_u(func, u):
    return [func((i, u)) for i in range(len(list(u)))]


def eval_AtA_times_u(u):
    return eval_times_u(part_At_times_u, eval_times_u(part_A_times_u, u))


def part_A_times_u(i_u):
    i, u = i_u
    partial_sum = 0
    for j, u_j in enumerate(u):
        partial_sum += eval_A(i, j) * u_j
    return partial_sum


def part_At_times_u(i_u):
    i, u = i_u
    partial_sum = 0
    for j, u_j in enumerate(u):
        partial_sum += eval_A(j, i) * u_j
    return partial_sum


def bench_spectral_norm(loops):
    range_it = range(loops)
    t0 = pyperf.perf_counter()

    for _ in range_it:
        u = [1] * DEFAULT_N

        for dummy in range(10):
            v = eval_AtA_times_u(u)
            u = eval_AtA_times_u(v)

        vBv = vv = 0

        for ue, ve in zip(u, v):
            vBv += ue * ve
            vv += ve * ve

    return pyperf.perf_counter() - t0


if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata['description'] = (
        'MathWorld: "Hundred-Dollar, Hundred-Digit Challenge Problems", '
        'Challenge #3.')
    runner.bench_time_func('spectral_norm', bench_spectral_norm)



import signal
import subprocess
from multiprocessing import Pool
from pathlib import Path

"""
Parallel backfilling helper for pyperformance runs on isolated CPUs.

Reads `sha=branch` pairs from backfill_shas.txt, invokes run-pyperformance.sh
for each revision, and lets that wrapper pin the workload to an isolated CPU,
materialize benchmark.conf, build CPython, and upload results to
speed.python.org. Stdout/stderr for each revision are captured under
output/<branch>-<sha>.(out|err).
"""


def get_revisions() -> tuple[str, str]:
    revisions = []
    with open("backfill_shas.txt", "r") as f:
        for line in f:
            sha, branch = line.split("=")
            revisions.append((sha, branch.rstrip()))
    return revisions


def run_pyperformance(revision):
    sha, branch = revision
    print(f"Running run-pyperformance.sh with sha: {sha}, branch: {branch}")
    output_dir = Path("output")
    output_dir.mkdir(parents=True, exist_ok=True)
    out_file = output_dir / f"{branch}-{sha}.out"
    err_file = output_dir / f"{branch}-{sha}.err"
    with open(out_file, "w") as output, open(err_file, "w") as error:
        subprocess.run(
            [
                "./run-pyperformance.sh",
                "-x",
                "--",
                "compile",
                "benchmark.conf",
                sha,
                branch,
            ],
            stdout=output,
            stderr=error,
        )


if __name__ == "__main__":
    original_sigint_handler = signal.signal(signal.SIGINT, signal.SIG_IGN)
    signal.signal(signal.SIGINT, original_sigint_handler)
    with Pool(8) as pool:
        res = pool.map_async(run_pyperformance, get_revisions())
        # Without the timeout this blocking call ignores all signals.
        res.get(86400)



import functools

import pyperf

from chameleon import PageTemplate


BIGTABLE_ZPT = """\
<table xmlns="http://www.w3.org/1999/xhtml"
xmlns:tal="http://xml.zope.org/namespaces/tal">
<tr tal:repeat="row python: options['table']">
<td tal:repeat="c python: row.values()">
<span tal:define="d python: c + 1"
tal:attributes="class python: 'column-' + str(d)"
tal:content="python: d" />
</td>
</tr>
</table>"""


def main():
    runner = pyperf.Runner()
    runner.metadata['description'] = "Chameleon template"

    tmpl = PageTemplate(BIGTABLE_ZPT)
    table = [dict(a=1, b=2, c=3, d=4, e=5, f=6, g=7, h=8, i=9, j=10)
             for x in range(500)]
    options = {'table': table}

    func = functools.partial(tmpl, options=options)
    runner.bench_func('chameleon', func)


if __name__ == '__main__':
    main()



__all__ = [
    "Benchmark",
    "BenchmarkSpec",
    "check_name",
    "parse_benchmark",
]


import os
import os.path
import sys
from collections import namedtuple

import pyperf
from packaging.specifiers import SpecifierSet

from . import _benchmark_metadata, _utils


def check_name(name):
    _utils.check_name("_" + name)


def parse_benchmark(entry, *, fail=True):
    name = entry
    version = None
    origin = None
    metafile = None

    if not f"_{name}".isidentifier():
        if not fail:
            return None
        raise ValueError(f"unsupported benchmark name in {entry!r}")

    bench = BenchmarkSpec(name, version, origin)
    return bench, metafile


class BenchmarkSpec(namedtuple("BenchmarkSpec", "name version origin")):
    __slots__ = ()

    @classmethod
    def from_raw(cls, raw):
        if isinstance(raw, BenchmarkSpec):
            return raw, None
        elif isinstance(raw, str):
            return parse_benchmark(raw)
        else:
            raise ValueError(f"unsupported raw spec {raw!r}")

    def __new__(cls, name, version=None, origin=None):
        self = super().__new__(cls, name, version or None, origin or None)
        return self


class Benchmark:
    _metadata = None

    def __init__(self, spec, metafile):
        spec, _metafile = BenchmarkSpec.from_raw(spec)
        if not metafile:
            if not _metafile:
                raise ValueError(f"missing metafile for {spec!r}")
            metafile = _metafile

        self.spec = spec
        self.metafile = metafile

    def __repr__(self):
        return f"{type(self).__name__}(spec={self.spec!r}, metafile={self.metafile!r})"

    def __hash__(self):
        return hash(self.spec)

    def __eq__(self, other):
        try:
            other_spec = other.spec
        except AttributeError:
            return NotImplemented
        return self.spec == other_spec

    def __gt__(self, other):
        try:
            other_spec = other.spec
        except AttributeError:
            return NotImplemented
        return self.spec > other_spec

    # __getattr__() gets weird when AttributeError comes out of
    # properties so we spell out all the aliased attributes.

    @property
    def name(self):
        return self.spec.name

    @property
    def version(self):
        version = self.spec.version
        if version is None:
            version = self._get_metadata_value("version", None)
        return version

    @property
    def origin(self):
        return self.spec.origin

    def _get_rootdir(self):
        try:
            return self._rootdir
        except AttributeError:
            script = self.runscript
            self._rootdir = os.path.dirname(script) if script else None
            return self._rootdir

    def _init_metadata(self):
        # assert self._metadata is None
        defaults = {
            "name": self.spec.name,
            "version": self.spec.version,
        }
        self._metadata, _ = _benchmark_metadata.load_metadata(
            self.metafile,
            defaults,
        )

    def _get_metadata_value(self, key, default):
        try:
            return self._metadata[key]
        except TypeError:
            if self._metadata is not None:
                raise  # re-raise
            self._init_metadata()
        except KeyError:
            pass
        return self._metadata.setdefault(key, default)

    @property
    def tags(self):
        return self._get_metadata_value("tags", [])

    @property
    def datadir(self):
        return self._get_metadata_value("datadir", None)

    @property
    def requirements_lockfile(self):
        try:
            return self._lockfile
        except AttributeError:
            lockfile = self._get_metadata_value("requirements_lockfile", None)
            if not lockfile:
                rootdir = self._get_rootdir()
                if rootdir:
                    lockfile = os.path.join(rootdir, "requirements.txt")
            self._lockfile = lockfile
            return self._lockfile

    @property
    def runscript(self):
        return self._get_metadata_value("runscript", None)

    @property
    def extra_opts(self):
        return self._get_metadata_value("extra_opts", ())

    @property
    def python(self):
        return SpecifierSet(self._get_metadata_value("python", ""))

    # Other metadata keys:
    # * base
    # * dependencies
    # * requirements

    def run(
        self,
        python,
        runid=None,
        pyperf_opts=None,
        *,
        venv=None,
        verbose=False,
    ):
        if venv and python == sys.executable:
            python = venv.python

        if not runid:
            from .run import get_run_id

            runid = get_run_id(python, self)

        runscript = self.runscript
        bench = _run_perf_script(
            python,
            runscript,
            runid,
            extra_opts=self.extra_opts,
            pyperf_opts=pyperf_opts,
            verbose=verbose,
        )

        return bench


#######################################
# internal implementation


def _run_perf_script(
    python,
    runscript,
    runid,
    *,
    extra_opts=None,
    pyperf_opts=None,
    verbose=False,
):
    if not runscript:
        raise ValueError("missing runscript")
    if not isinstance(runscript, str):
        raise TypeError(f"runscript must be a string, got {runscript!r}")

    with _utils.temporary_file() as tmp:
        opts = [
            *(extra_opts or ()),
            *(pyperf_opts or ()),
            "--output",
            tmp,
        ]
        if pyperf_opts and "--copy-env" in pyperf_opts:
            argv, env = _prep_cmd(python, runscript, opts, runid, lambda name: None)
        else:
            opts, inherit_envvar = _resolve_restricted_opts(opts)
            argv, env = _prep_cmd(python, runscript, opts, runid, inherit_envvar)
        hide_stderr = not verbose
        ec, _, stderr = _utils.run_cmd(
            argv,
            env=env,
            capture="stderr" if hide_stderr else None,
        )
        if ec != 0:
            if hide_stderr:
                sys.stderr.flush()
                sys.stderr.write(stderr)
                sys.stderr.flush()
            # pyperf returns exit code 124 if the benchmark execution times out
            if ec == 124:
                raise TimeoutError("Benchmark timed out")
            else:
                raise RuntimeError("Benchmark died")
        return pyperf.BenchmarkSuite.load(tmp)


def _prep_cmd(python, script, opts, runid, on_set_envvar=None):
    # Populate the environment variables.
    env = dict(os.environ)

    def set_envvar(name, value):
        env[name] = value
        if on_set_envvar is not None:
            on_set_envvar(name)

    # on_set_envvar() may update "opts" so all calls to set_envvar()
    # must happen before building argv.
    set_envvar("PYPERFORMANCE_RUNID", str(runid))

    # Build argv.
    argv = [
        python,
        "-u",
        script,
        *(opts or ()),
    ]

    return argv, env


def _resolve_restricted_opts(opts):
    # Deal with --inherit-environ.
    FLAG = "--inherit-environ"
    resolved = []
    idx = None
    for i, opt in enumerate(opts):
        if opt.startswith(FLAG + "="):
            idx = i + 1
            resolved.append(FLAG)
            resolved.append(opt.partition("=")[-1])
            resolved.extend(opts[idx:])
            break
        elif opt == FLAG:
            idx = i + 1
            resolved.append(FLAG)
            resolved.append(opts[idx])
            resolved.extend(opts[idx + 1 :])
            break
        else:
            resolved.append(opt)
    else:
        resolved.extend(["--inherit-environ", ""])
        idx = len(resolved) - 1
    inherited = set(resolved[idx].replace(",", " ").split())

    def inherit_env_var(name):
        inherited.add(name)
        resolved[idx] = ",".join(inherited)

    return resolved, inherit_env_var


def _insert_on_PYTHONPATH(entry, env):
    PYTHONPATH = env.get("PYTHONPATH", "").split(os.pathsep)
    PYTHONPATH.insert(0, entry)
    env["PYTHONPATH"] = os.pathsep.join(PYTHONPATH)



import os
import os.path
import sys

from . import _utils

GET_PIP_URL = "https://bootstrap.pypa.io/get-pip.py"
# pip 6 is the first version supporting environment markers
MIN_PIP = "6.0"
OLD_SETUPTOOLS = "18.5"


def get_pkg_name(req):
    """Return the name of the package in the given requirement text."""
    # strip env markers
    req = req.partition(";")[0]
    # strip version
    req = req.partition("==")[0]
    req = req.partition(">=")[0]
    return req


def run_pip(cmd, *args, **kwargs):
    """Return the result of running pip with the given args."""
    return _utils.run_python("-m", "pip", cmd, *args, **kwargs)


def is_pip_installed(python, *, env=None):
    """Return True if pip is installed on the given Python executable."""
    ec, _, _ = run_pip(
        "--version",
        python=python,
        env=env,
        capture=True,
        verbose=False,
    )
    return ec == 0


def install_pip(
    python=sys.executable,
    *,
    info=None,
    downloaddir=None,
    env=None,
    upgrade=True,
    **kwargs,
):
    """Install pip on the given Python executable."""
    if not python:
        python = getattr(info, "executable", None) or sys.executable

    # python -m ensurepip
    args = ["-m", "ensurepip", "-v"]  # --verbose
    if upgrade:
        args.append("-U")  # --upgrade
    res = _utils.run_python(*args, python=python, **kwargs)
    ec, _, _ = res
    if ec == 0 and is_pip_installed(python, env=env):
        return res

    ##############################
    # Fall back to get-pip.py.

    if not downloaddir:
        downloaddir = "."
    os.makedirs(downloaddir, exist_ok=True)

    # download get-pip.py
    filename = os.path.join(downloaddir, "get-pip.py")
    if not os.path.exists(filename):
        print("Download %s into %s" % (GET_PIP_URL, filename))
        _utils.download(GET_PIP_URL, filename)

    # python get-pip.py
    argv = [python, "-u", filename]
    res = _utils.run_cmd(argv, env=env)
    ec, _, _ = res
    if ec != 0:
        # get-pip.py was maybe not properly downloaded: remove it to
        # download it again next time
        os.unlink(filename)
    return res


def upgrade_pip(
    python=sys.executable,
    *,
    info=None,
    installer=False,
    **kwargs,
):
    """Upgrade pip on the given Python to the latest version."""
    if not python:
        python = getattr(info, "executable", None) or sys.executable

    # pip 6 is the first version supporting environment markers
    reqs = [f"pip>={MIN_PIP}"]
    res = install_requirements(*reqs, python=python, upgrade=True, **kwargs)
    ec, _, _ = res
    if ec != 0:
        return res

    if installer:
        # Upgrade installer dependencies (setuptools, ...)
        res = ensure_installer(python, upgrade=True, **kwargs)
    return res


def ensure_installer(python=sys.executable, **kwargs):
    reqs = [
        f"setuptools>={OLD_SETUPTOOLS}",
        # install wheel so pip can cache binary wheel packages locally,
        # and install prebuilt wheel packages from PyPI.
        "wheel",
    ]
    return install_requirements(*reqs, python=python, **kwargs)


def install_requirements(reqs, *extra, upgrade=True, **kwargs):
    """Install the given packages from PyPI."""
    args = []
    if upgrade:
        args.append("-U")  # --upgrade
    for reqs in [reqs, *extra]:
        if os.path.isfile(reqs) and reqs.endswith(".txt"):
            args.append("-r")  # --requirement
        args.append(reqs)
    return run_pip("install", *args, **kwargs)


def install_editable(projectroot, **kwargs):
    """Install the given project as an "editable" install."""
    return run_pip("install", "-e", projectroot, **kwargs)



"""
Benchmark to measure performance of the python builtin method copy.deepcopy

Performance is tested on a nested dictionary and a dataclass

Author: Pieter Eendebak

"""
import copy
import pyperf
from dataclasses import dataclass


@dataclass
class A:
    string: str
    lst: list
    boolean: bool


def benchmark_reduce(n):
    """ Benchmark where the __reduce__ functionality is used """
    class C(object):
        def __init__(self):
            self.a = 1
            self.b = 2

        def __reduce__(self):
            return (C, (), self.__dict__)

        def __setstate__(self, state):
            self.__dict__.update(state)
    c = C()

    t0 = pyperf.perf_counter()
    for ii in range(n):
        _ = copy.deepcopy(c)
    dt = pyperf.perf_counter() - t0
    return dt


def benchmark_memo(n):
    """ Benchmark where the memo functionality is used """
    A = [1] * 100
    data = {'a': (A, A, A), 'b': [A] * 100}

    t0 = pyperf.perf_counter()
    for ii in range(n):
        _ = copy.deepcopy(data)
    dt = pyperf.perf_counter() - t0
    return dt


def benchmark(n):
    """ Benchmark on some standard data types """
    a = {
        'list': [1, 2, 3, 43],
        't': (1 ,2, 3),
        'str': 'hello',
        'subdict': {'a': True}
    }
    dc = A('hello', [1, 2, 3], True)

    dt = 0
    for ii in range(n):
        for jj in range(30):
            t0 = pyperf.perf_counter()
            _ = copy.deepcopy(a)
            dt += pyperf.perf_counter() - t0
        for s in ['red', 'blue', 'green']:
            dc.string = s
            for kk in range(5):
                dc.lst[0] = kk
                for b in [True, False]:
                    dc.boolean = b
                    t0 = pyperf.perf_counter()
                    _ = copy.deepcopy(dc)
                    dt += pyperf.perf_counter() - t0
    return dt


if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata['description'] = "deepcopy benchmark"

    runner.bench_time_func('deepcopy', benchmark)
    runner.bench_time_func('deepcopy_reduce', benchmark_reduce)
    runner.bench_time_func('deepcopy_memo', benchmark_memo)



import configparser
import datetime
import errno
import json
import logging
import math
import os
import os.path
import re
import shlex
import statistics
import subprocess
import sys
import time
from urllib.error import HTTPError
from urllib.parse import urlencode
from urllib.request import urlopen

import pyperf

import pyperformance
from pyperformance import _pip, _utils

GIT = True
DEFAULT_BRANCH = "master" if GIT else "default"
LOG_FORMAT = "%(asctime)-15s: %(message)s"

EXIT_ALREADY_EXIST = 10
EXIT_COMPILE_ERROR = 11
EXIT_VENV_ERROR = 11
EXIT_BENCH_ERROR = 12


def parse_date(text):
    def replace_timezone(regs):
        text = regs.group(0)
        return text[:2] + text[3:]

    # replace '+01:00' with '+0100'
    text2 = re.sub(r"[0-9]{2}:[0-9]{2}$", replace_timezone, text)

    # ISO 8601 with timezone: '2017-03-30T19:12:18+00:00'
    return datetime.datetime.strptime(text2, "%Y-%m-%dT%H:%M:%S%z")


class Task(object):
    def __init__(self, app, cwd):
        self.app = app
        self.cwd = cwd

    def get_output_nocheck(self, *cmd):
        return self.app.get_output_nocheck(*cmd, cwd=self.cwd)

    def get_output(self, *cmd):
        return self.app.get_output(*cmd, cwd=self.cwd)

    def run_nocheck(self, *cmd, **kw):
        return self.app.run_nocheck(*cmd, cwd=self.cwd, **kw)

    def run(self, *cmd, **kw):
        self.app.run(*cmd, cwd=self.cwd, **kw)


class Repository(Task):
    def __init__(self, app, path):
        super().__init__(app, path)
        self.logger = app.logger
        self.app = app
        self.conf = app.conf

    def fetch(self):
        if GIT:
            self.run("git", "fetch")
        else:
            self.run("hg", "pull")

    def parse_revision(self, revision):
        branch_rev = "%s/%s" % (self.conf.git_remote, revision)

        exitcode, stdout = self.get_output_nocheck(
            "git", "rev-parse", "--verify", branch_rev
        )
        if not exitcode:
            return (True, branch_rev, stdout)

        exitcode, stdout = self.get_output_nocheck(
            "git", "rev-parse", "--verify", revision
        )
        if not exitcode and stdout.startswith(revision):
            revision = stdout
            return (False, revision, revision)

        self.logger.error("ERROR: unable to parse revision %r" % (revision,))
        sys.exit(1)

    def checkout(self, revision):
        if GIT:
            # remove all untracked files
            self.run("git", "clean", "-fdx")

            # checkout to requested revision
            self.run("git", "reset", "--hard", "HEAD")
            self.run("git", "checkout", revision)

            # remove all untracked files
            self.run("git", "clean", "-fdx")
        else:
            self.run("hg", "up", "--clean", "-r", revision)
            # FIXME: run hg purge?

    def get_revision_info(self, revision):
        if GIT:
            cmd = ["git", "show", "-s", "--pretty=format:%H|%ci", "%s^!" % revision]
        else:
            cmd = ["hg", "log", "--template", "{node}|{date|isodate}", "-r", revision]
        stdout = self.get_output(*cmd)
        if GIT:
            node, date = stdout.split("|")
            date = datetime.datetime.strptime(date, "%Y-%m-%d %H:%M:%S %z")
            # convert local date to UTC
            date = (date - date.utcoffset()).replace(tzinfo=datetime.timezone.utc)
        else:
            node, date = stdout.split("|")
            date = datetime.datetime.strptime(date[:16], "%Y-%m-%d %H:%M")
        return (node, date)


class Application(object):
    def __init__(self, conf, options):
        self.conf = conf
        self.options = options
        logging.basicConfig(format=LOG_FORMAT)
        self.logger = logging.getLogger()
        self.log_filename = None

    def setup_log(self, prefix):
        prefix = re.sub("[^A-Za-z0-9_-]+", "_", prefix)
        prefix = re.sub("_+", "_", prefix)

        date = datetime.datetime.now()
        date = date.strftime("%Y-%m-%d_%H-%M-%S.log")
        filename = "%s-%s" % (prefix, date)
        self.log_filename = os.path.join(self.conf.directory, filename)

        log = self.log_filename
        if os.path.exists(log):
            self.logger.error("ERROR: Log file %s already exists" % log)
            sys.exit(1)
        self.safe_makedirs(os.path.dirname(log))

        handler = logging.FileHandler(log)
        formatter = logging.Formatter(LOG_FORMAT)
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)

    def create_subprocess(self, cmd, **kwargs):
        self.logger.error("+ %s" % " ".join(map(shlex.quote, cmd)))
        return subprocess.Popen(cmd, **kwargs)

    def run_nocheck(self, *cmd, stdin_filename=None, **kwargs):
        if stdin_filename:
            stdin_file = open(stdin_filename, "rb", 0)
            kwargs["stdin"] = stdin_file.fileno()
        else:
            stdin_file = None
        log_stdout = kwargs.pop("log_stdout", True)

        if log_stdout:
            kwargs["stdout"] = subprocess.PIPE
            kwargs["stderr"] = subprocess.STDOUT
            kwargs["universal_newlines"] = True

        try:
            proc = self.create_subprocess(cmd, **kwargs)

            # FIXME: support Python 2?
            with proc:
                if log_stdout:
                    for line in proc.stdout:
                        line = line.rstrip()
                        self.logger.error(line)
                exitcode = proc.wait()
        finally:
            if stdin_file is not None:
                stdin_file.close()

        if exitcode:
            cmd_str = " ".join(map(shlex.quote, cmd))
            self.logger.error(
                "Command %s failed with exit code %s" % (cmd_str, exitcode)
            )

        return exitcode

    def run(self, *cmd, **kw):
        exitcode = self.run_nocheck(*cmd, **kw)
        if exitcode:
            sys.exit(exitcode)

    def get_output_nocheck(self, *cmd, **kwargs):
        proc = self.create_subprocess(
            cmd, stdout=subprocess.PIPE, universal_newlines=True, **kwargs
        )
        # FIXME: support Python 2?
        with proc:
            stdout = proc.communicate()[0]

        stdout = stdout.rstrip()

        exitcode = proc.wait()
        if exitcode:
            cmd_str = " ".join(map(shlex.quote, cmd))
            self.logger.error(
                "Command %s failed with exit code %s" % (cmd_str, exitcode)
            )

        return (exitcode, stdout)

    def get_output(self, *cmd, **kwargs):
        exitcode, stdout = self.get_output_nocheck(*cmd, **kwargs)
        if exitcode:
            for line in stdout.splitlines():
                self.logger.error(line)
            sys.exit(exitcode)

        return stdout

    def safe_makedirs(self, directory):
        try:
            os.makedirs(directory)
        except OSError as exc:
            if exc.errno != errno.EEXIST:
                raise


def resolve_python(prefix, builddir, *, fallback=True):
    if sys.platform in ("darwin", "win32"):
        program_ext = ".exe"
    else:
        program_ext = ""

    if prefix:
        if sys.platform == "darwin":
            program_ext = ""
        program = os.path.join(prefix, "bin", "python3" + program_ext)
        exists = os.path.exists(program)
        if not exists and fallback:
            program2 = os.path.join(prefix, "bin", "python" + program_ext)
            if os.path.exists(program2):
                program = program2
                exists = True
    else:
        assert builddir
        program = os.path.join(builddir, "python" + program_ext)
        exists = os.path.exists(program)
    return program, exists


class Python(Task):
    def __init__(self, app, conf):
        super().__init__(app, conf.build_dir)
        self.app = app
        self.branch = app.branch
        self.conf = conf
        self.logger = app.logger
        self.program = None
        self.hexversion = None

    def patch(self, filename):
        if not filename:
            return

        self.logger.error(
            "Apply patch %s in %s (revision %s)"
            % (filename, self.conf.repo_dir, self.app.revision)
        )
        self.app.run("patch", "-p1", cwd=self.conf.repo_dir, stdin_filename=filename)

    def compile(self):
        build_dir = self.conf.build_dir

        _utils.safe_rmtree(build_dir)
        self.app.safe_makedirs(build_dir)

        config_args = []
        if self.branch.startswith("2.") and not _utils.MS_WINDOWS:
            # On Python 2, use UCS-4 for Unicode on all platforms, except
            # on Windows which uses UTF-16 because of its 16-bit wchar_t
            config_args.append("--enable-unicode=ucs4")
        if self.conf.prefix:
            config_args.extend(("--prefix", self.conf.prefix))
        if self.conf.debug:
            config_args.append("--with-pydebug")
        elif self.conf.lto:
            config_args.append("--with-lto")
        if self.conf.jit:
            config_args.append(f"--enable-experimental-jit={self.conf.jit}")
        if self.conf.pkg_only:
            config_args.extend(self.get_package_only_flags())
        if self.conf.debug:
            config_args.append("CFLAGS=-O0")
        configure = os.path.join(self.conf.repo_dir, "configure")
        self.run(configure, *config_args)

        argv = ["make"]
        if self.conf.pgo:
            # FIXME: use taskset (isolated CPUs) for PGO?
            argv.append("profile-opt")
        if self.conf.jobs:
            argv.append("-j%d" % self.conf.jobs)
        self.run(*argv)

    def install_python(self):
        program, _ = resolve_python(
            self.conf.prefix if self.conf.install else None,
            self.conf.build_dir,
        )
        if self.conf.install:
            program, _ = resolve_python(self.conf.prefix, self.conf.build_dir)
            _utils.safe_rmtree(self.conf.prefix)
            self.app.safe_makedirs(self.conf.prefix)
            self.run("make", "install")
        else:
            program, _ = resolve_python(None, self.conf.build_dir)
        # else don't install: run python from the compilation directory
        self.program = program

    def get_version(self):
        # Dump the Python version
        self.logger.error("Installed Python version:")
        self.run(self.program, "--version")

        # Get the Python version
        code = "import sys; print(sys.hexversion)"
        stdout = self.get_output(self.program, "-c", code)
        self.hexversion = int(stdout)
        self.logger.error("Python hexversion: %x" % self.hexversion)

    def get_package_only_flags(self):
        arguments = []
        extra_paths = []

        for pkg in self.conf.pkg_only:
            prefix = self.get_package_prefix(pkg)
            if pkg == "openssl":
                arguments.append("--with-openssl=" + prefix)
            else:
                extra_paths.append(prefix)

        if extra_paths:
            # Flags are one CLI arg each and do not need quotes.
            ps = ["-I%s/include" % p for p in extra_paths]
            arguments.append("CFLAGS=%s" % " ".join(ps))
            ps = ["-L%s/lib" % p for p in extra_paths]
            arguments.append("LDFLAGS=%s" % " ".join(ps))
        return arguments

    def get_package_prefix(self, name):
        if sys.platform == "darwin":
            cmd = ["brew", "--prefix", name]
        else:
            self.logger.error(
                "ERROR: package-only libraries are not supported on %s" % sys.platform
            )
            sys.exit(1)

        stdout = self.get_output(*cmd)
        self.logger.error("Package prefix for %s is '%s'" % (name, stdout))
        return stdout

    def download(self, url, filename):
        self.logger.error("Download %s into %s" % (url, filename))
        _utils.download(url, filename)

    def _install_pip(self):
        # is pip already installed and working?
        exitcode = self.run_nocheck(self.program, "-u", "-m", "pip", "--version")
        if not exitcode:
            # Upgrade pip
            self.run(self.program, "-u", "-m", "pip", "install", "-U", "pip")
            return

        # pip is missing (or broken?): install it
        filename = os.path.join(self.conf.directory, "get-pip.py")
        if not os.path.exists(filename):
            self.download(_pip.GET_PIP_URL, filename)

        # Install pip
        self.run(self.program, "-u", filename)

    def install_pip(self):
        self._install_pip()

        # Dump the pip version
        self.run(self.program, "-u", "-m", "pip", "--version")

    def install_performance(self):
        cmd = [self.program, "-u", "-m", "pip", "install"]

        if pyperformance.is_dev():
            cmd.extend(["-e", os.path.dirname(pyperformance.PKG_ROOT)])
        else:
            version = pyperformance.__version__
            cmd.append("pyperformance==%s" % version)

        self.run(*cmd)

    def compile_install(self):
        self.compile()
        self.install_python()
        self.get_version()
        self.install_pip()
        self.install_performance()


class BenchmarkRevision(Application):
    _dryrun = False

    def __init__(
        self,
        conf,
        revision,
        branch=None,
        patch=None,
        setup_log=True,
        filename=None,
        commit_date=None,
        options=None,
    ):
        super().__init__(conf, options)
        self.patch = patch
        self.exitcode = 0
        self.uploaded = False

        if setup_log:
            if branch:
                prefix = "compile-%s-%s" % (branch, revision)
            else:
                prefix = "compile-%s" % revision
            self.setup_log(prefix)

        if filename is None:
            self.repository = Repository(self, conf.repo_dir)
            self.init_revision(revision, branch)
        else:
            # path used by cmd_upload()
            self.repository = None
            self.filename = filename
            self.revision = revision
            self.branch = branch
            self.commit_date = commit_date
            self.logger.error(
                "Commit: branch=%s, revision=%s" % (self.branch, self.revision)
            )

        self.upload_filename = os.path.join(
            self.conf.uploaded_json_dir, os.path.basename(self.filename)
        )

    def init_revision(self, revision, branch=None):
        if self.conf.update:
            self.repository.fetch()

        if branch:
            is_branch, rev_name, full_revision = self.repository.parse_revision(branch)
            if not is_branch:
                self.logger.error("ERROR: %r is not a Git branch" % self.branch)
                sys.exit(1)
            self.branch = branch
        else:
            self.branch = None

        is_branch, rev_name, full_revision = self.repository.parse_revision(revision)
        if is_branch:
            if self.branch and revision != self.branch:
                raise ValueError(
                    "inconsistenct branches: "
                    "revision=%r, branch=%r" % (revision, branch)
                )
            self.branch = revision
        elif not self.branch:
            self.branch = DEFAULT_BRANCH

        self.revision, date = self.repository.get_revision_info(rev_name)
        self.logger.error(
            "Commit: branch=%s, revision=%s, date=%s"
            % (self.branch, self.revision, date)
        )
        self.commit_date = date

        date = date.strftime("%Y-%m-%d_%H-%M")

        filename = "%s-%s-%s" % (date, self.branch, self.revision[:12])
        if self.patch:
            patch = os.path.basename(self.patch)
            patch = os.path.splitext(patch)[0]
            filename = "%s-patch-%s" % (filename, patch)
        filename = filename + ".json.gz"
        if self.patch:
            self.filename = os.path.join(self.conf.json_patch_dir, filename)
        else:
            self.filename = os.path.join(self.conf.json_dir, filename)

    def compile_install(self):
        self.repository.checkout(self.revision)

        # First: remove everything
        _utils.safe_rmtree(self.conf.build_dir)
        _utils.safe_rmtree(self.conf.prefix)

        self.python.patch(self.patch)
        self.python.compile_install()

    def create_venv(self):
        # Create venv
        python = self.python.program
        if self._dryrun:
            program, exists = resolve_python(
                self.conf.prefix if self.conf.install else None,
                self.conf.build_dir,
            )
            if not python or not exists:
                python = sys.executable
        cmd = [
            python,
            "-u",
            "-m",
            "pyperformance",
            "venv",
            "recreate",
            "--venv",
            self.conf.venv,
            "--benchmarks",
            "<NONE>",
        ]
        if self.options.inherit_environ:
            cmd.append("--inherit-environ=%s" % ",".join(self.options.inherit_environ))
        exitcode = self.run_nocheck(*cmd)
        if exitcode:
            sys.exit(EXIT_VENV_ERROR)
        binname = "Scripts" if os.name == "nt" else "bin"
        base = os.path.basename(python)
        return os.path.join(self.conf.venv, binname, base)

    def run_benchmark(self, python=None):
        self.safe_makedirs(os.path.dirname(self.filename))
        if not python:
            python = self.python.program
        if self._dryrun:
            python = sys.executable
        cmd = [
            python,
            "-u",
            "-m",
            "pyperformance",
            "run",
            "--verbose",
            "--output",
            self.filename,
        ]
        if self.options.inherit_environ:
            cmd.append("--inherit-environ=%s" % ",".join(self.options.inherit_environ))
        if self.conf.manifest:
            cmd.extend(("--manifest", self.conf.manifest))
        if self.conf.benchmarks:
            cmd.append("--benchmarks=%s" % self.conf.benchmarks)
        if self.conf.affinity:
            cmd.extend(("--affinity", self.conf.affinity))
        if self.conf.debug:
            cmd.append("--debug-single-value")
        if self.conf.same_loops:
            cmd.append("--same_loops=%s" % self.conf.same_loops)
        if self.conf.rigorous:
            cmd.append("--rigorous")
        exitcode = self.run_nocheck(*cmd)

        if os.path.exists(self.filename):
            self.update_metadata()

        return bool(exitcode)

    def update_metadata(self):
        metadata = {
            "commit_id": self.revision,
            "commit_branch": self.branch,
            "commit_date": self.commit_date.isoformat(),
        }
        if self.patch:
            metadata["patch_file"] = self.patch

        suite = pyperf.BenchmarkSuite.load(self.filename)
        for bench in suite:
            bench.update_metadata(metadata)
        suite.dump(self.filename, replace=True)

    def encode_benchmark(self, bench):
        data = {}
        data["environment"] = self.conf.environment
        data["project"] = self.conf.project
        data["branch"] = self.branch
        data["benchmark"] = bench.get_name()
        # Other benchmark metadata:
        # - description
        # - units="seconds", units_title="Time", lessisbetter=True
        data["commitid"] = self.revision
        data["revision_date"] = self.commit_date.isoformat()
        data["executable"] = self.conf.executable
        data["result_value"] = bench.mean()
        # Other result metadata: result_date
        if bench.get_nvalue() == 1:
            data["std_dev"] = 0
        else:
            data["std_dev"] = bench.stdev()
        values = bench.get_values()
        data["min"] = min(values)
        data["max"] = max(values)
        # Other stats metadata: q1, q3
        return data

    def upload(self):
        if self.uploaded:
            raise Exception("already uploaded")

        if self.filename == self.upload_filename:
            self.logger.error("ERROR: %s was already uploaded!" % self.filename)
            sys.exit(1)

        if os.path.exists(self.upload_filename):
            self.logger.error(
                "ERROR: cannot upload, %s file ready exists!" % self.upload_filename
            )
            sys.exit(1)

        self.safe_makedirs(self.conf.uploaded_json_dir)

        suite = pyperf.BenchmarkSuite.load(self.filename)
        data = [self.encode_benchmark(bench) for bench in suite]
        data = dict(json=json.dumps(data))

        url = self.conf.url
        if not url.endswith("/"):
            url += "/"
        url += "result/add/json/"
        self.logger.error("Upload %s benchmarks to %s" % (len(suite), url))

        try:
            response = urlopen(data=urlencode(data).encode("utf-8"), url=url)
            body = response.read()
            response.close()
        except HTTPError as err:
            self.logger.error("HTTP Error: %s" % err)
            errmsg = err.read().decode("utf8")
            self.logger.error(errmsg)
            err.close()
            return

        self.logger.error('Response: "%s"' % body.decode("utf-8"))

        self.logger.error("Move %s to %s" % (self.filename, self.upload_filename))
        os.rename(self.filename, self.upload_filename)

        self.uploaded = True

    def perf_system_tune(self):
        pythonpath = os.environ.get("PYTHONPATH")
        args = ["-m", "pyperf", "system", "tune"]
        if self.conf.affinity:
            args.extend(("--affinity", self.conf.affinity))
        if pythonpath:
            cmd = "PYTHONPATH=%s %s %s" % (
                shlex.quote(pythonpath),
                shlex.quote(sys.executable),
                " ".join(args),
            )
            self.run("sudo", "bash", "-c", cmd)
        else:
            self.run("sudo", sys.executable, *args)

    def prepare(self):
        self.logger.error(
            "Compile and benchmarks Python rev %s (branch %s)"
            % (self.revision, self.branch)
        )
        self.logger.error("")

        if os.path.exists(self.filename):
            filename = self.filename
        elif self.conf.upload and os.path.exists(self.upload_filename):
            filename = self.upload_filename
        else:
            filename = False
        if filename:
            # Benchmark already uploaded
            self.logger.error("JSON file %s already exists: do nothing" % filename)

            # Remove the log file
            if self.log_filename:
                self.logger.error("Remove log file %s" % self.log_filename)
                del self.logger.handlers[:]
                os.unlink(self.log_filename)

            sys.exit(EXIT_ALREADY_EXIST)

        if self.log_filename:
            self.logger.error("Write logs into %s" % self.log_filename)

        if self.patch and self.conf.upload:
            self.logger.error("Disable upload on patched Python")
            self.conf.upload = False

        if self.conf.debug and self.conf.upload:
            self.logger.error("Disable upload in debug mode")
            self.conf.upload = False

        if not self.conf.install and self.conf.upload:
            self.logger.error("Disable upload if Python is not installed")
            self.conf.upload = False

        if self.conf.system_tune:
            self.perf_system_tune()

    def compile_bench(self):
        self.python = Python(self, self.conf)

        if not self._dryrun:
            try:
                self.compile_install()
            except SystemExit:
                sys.exit(EXIT_COMPILE_ERROR)

        if self.conf.venv:
            python = self.create_venv()
        else:
            python = None

        failed = self.run_benchmark(python)
        if self.conf.upload:
            self.upload()
        return failed

    def main(self):
        self.start = time.monotonic()

        self.prepare()
        failed = self.compile_bench()

        dt = time.monotonic() - self.start
        dt = datetime.timedelta(seconds=dt)
        self.logger.error("Benchmark completed in %s" % dt)

        if self.uploaded:
            self.logger.error(
                "Benchmark results uploaded and written into %s" % self.upload_filename
            )
        elif failed:
            self.logger.error(
                "Benchmark failed but results written into %s" % self.filename
            )
        else:
            self.logger.error("Benchmark result written into %s" % self.filename)

        if failed:
            sys.exit(EXIT_BENCH_ERROR)


class Configuration:
    pass


def parse_config(filename, command):
    parse_compile = False
    parse_compile_all = False
    if command == "compile_all":
        parse_compile = True
        parse_compile_all = True
    elif command == "compile":
        parse_compile = True
    else:
        assert command == "upload"

    conf = Configuration()
    cfgobj = configparser.ConfigParser()
    cfgobj.read(filename)

    def getstr(section, key, default=None):
        try:
            sectionobj = cfgobj[section]
            value = sectionobj[key]
        except KeyError:
            if default is None:
                raise
            return default

        # strip comments
        value = value.partition("#")[0]
        # strip spaces
        return value.strip()

    def getfile(section, key, default=None):
        value = getstr(section, key, default)
        if not value:
            return value
        value = os.path.expanduser(value)
        return value

    def getboolean(section, key, default):
        try:
            sectionobj = cfgobj[section]
            return sectionobj.getboolean(key, default)
        except KeyError:
            return default

    def getint(section, key, default=None):
        return int(getstr(section, key, default))

    # [config]
    conf.json_dir = getfile("config", "json_dir")
    conf.json_patch_dir = os.path.join(conf.json_dir, "patch")
    conf.uploaded_json_dir = os.path.join(conf.json_dir, "uploaded")
    conf.debug = getboolean("config", "debug", False)

    if parse_compile:
        # [scm]
        conf.repo_dir = getfile("scm", "repo_dir")
        conf.update = getboolean("scm", "update", True)
        conf.git_remote = getstr("config", "git_remote", default="remotes/origin")

        # [compile]
        conf.directory = getfile("compile", "bench_dir")
        conf.lto = getboolean("compile", "lto", True)
        conf.pgo = getboolean("compile", "pgo", True)
        conf.jit = getstr("compile", "jit", "")
        conf.install = getboolean("compile", "install", True)
        conf.pkg_only = getstr("compile", "pkg_only", "").split()
        try:
            conf.jobs = getint("compile", "jobs")
        except KeyError:
            conf.jobs = None

        # [run_benchmark]
        conf.system_tune = getboolean("run_benchmark", "system_tune", True)
        conf.manifest = getfile("run_benchmark", "manifest", default="")
        conf.benchmarks = getstr("run_benchmark", "benchmarks", default="")
        conf.affinity = getstr("run_benchmark", "affinity", default="")
        conf.upload = getboolean("run_benchmark", "upload", False)
        conf.same_loops = getfile("run_benchmark", "same_loops", default="")
        conf.rigorous = getboolean("run_benchmark", "rigorous", False)

        # paths
        conf.build_dir = os.path.join(conf.directory, "build")
        conf.prefix = os.path.join(conf.directory, "prefix")
        conf.venv = os.path.join(conf.directory, "venv")

        check_upload = conf.upload
    else:
        check_upload = True

    # [upload]
    UPLOAD_OPTIONS = ("url", "environment", "executable", "project")

    conf.url = getstr("upload", "url", default="")
    conf.executable = getstr("upload", "executable", default="")
    conf.project = getstr("upload", "project", default="")
    conf.environment = getstr("upload", "environment", default="")

    if check_upload and any(not getattr(conf, attr) for attr in UPLOAD_OPTIONS):
        print(
            "ERROR: Upload requires to set the following "
            "configuration option in the the [upload] section "
            "of %s:" % filename
        )
        for attr in UPLOAD_OPTIONS:
            text = "- %s" % attr
            if not getattr(conf, attr):
                text += " (not set)"
            print(text)
        sys.exit(1)

    if parse_compile_all:
        # [compile_all]
        conf.branches = getstr("compile_all", "branches", "").split()
        conf.revisions = []
        try:
            revisions = cfgobj.items("compile_all_revisions")
        except configparser.NoSectionError:
            pass
        else:
            for revision, name in revisions:
                # strip comments
                name = name.partition("#")[0]
                # strip spaces
                name = name.strip()
                conf.revisions.append((revision, name))

    # process config
    if conf.debug:
        conf.pgo = False
        conf.lto = False

    return conf


class BenchmarkAll(Application):
    def __init__(self, config_filename, options):
        config_filename = os.path.abspath(config_filename)
        conf = parse_config(config_filename, "compile_all")
        super().__init__(conf, options)
        self.config_filename = config_filename
        self.safe_makedirs(self.conf.directory)
        self.setup_log("compile_all")
        self.outputs = []
        self.skipped = []
        self.failed = []
        self.timings = []
        self.logger = logging.getLogger()

    def benchmark(self, revision, branch):
        if branch:
            key = "%s-%s" % (branch, revision)
        else:
            key = revision

        cmd = [
            sys.executable,
            "-m",
            "pyperformance",
            "compile",
            self.config_filename,
            revision,
            branch,
        ]
        if not self.conf.update:
            cmd.append("--no-update")
        if not self.conf.system_tune:
            cmd.append("--no-tune")

        self.start = time.monotonic()
        exitcode = self.run_nocheck(*cmd, log_stdout=False)
        dt = time.monotonic() - self.start

        if exitcode:
            self.logger.error("Benchmark exit code: %s" % exitcode)

        if exitcode == EXIT_ALREADY_EXIST:
            # Benchmark already uploaded
            self.skipped.append(key)
            return

        # compile, venv and bench errors occur after repository update
        # and system tune
        if exitcode >= EXIT_COMPILE_ERROR:
            if self.conf.system_tune:
                # only tune the system once
                self.conf.system_tune = False

            if self.conf.update:
                # Ony update the repository once
                self.conf.update = False

        if exitcode == 0 or exitcode == EXIT_BENCH_ERROR:
            self.outputs.append((key, exitcode == EXIT_BENCH_ERROR))
            self.timings.append(dt)
        else:
            self.failed.append(key)

    def report(self):
        for key in self.skipped:
            self.logger.error("Skipped: %s" % key)

        for key, success in self.outputs:
            if success:
                success_message = "All benchmarks succeeded"
            else:
                success_message = "Some benchmarks failed"
            self.logger.error("Tested: %s (%s)" % (key, success_message))

        for key in self.failed:
            text = "FAILED: %s" % key
            self.logger.error(text)

    def report_timings(self):
        def format_time(seconds):
            if seconds >= 100:
                return "%.0f min %.0f sec" % divmod(seconds, 60)
            else:
                return "%.0f sec" % math.ceil(seconds)

        self.logger.error("Timings:")
        self.logger.error("- min: %s" % format_time(min(self.timings)))
        text = "- avg: %s" % format_time(statistics.mean(self.timings))
        if len(self.timings) >= 2:
            stdev = statistics.stdev(self.timings)
            text = "%s -- std dev: %s" % (text, format_time(stdev))
        self.logger.error(text)
        self.logger.error("- max: %s" % format_time(max(self.timings)))

    def main(self):
        self.safe_makedirs(self.conf.directory)

        self.logger.error("Compile and benchmark all")
        if self.log_filename:
            self.logger.error("Write logs into %s" % self.log_filename)
        self.logger.error("Revisions: %r" % (self.conf.revisions,))
        self.logger.error("Branches: %r" % (self.conf.branches,))

        if not self.conf.revisions and not self.conf.branches:
            self.logger.error(
                "ERROR: no branches nor revisions configured for compile_all"
            )
            sys.exit(1)

        try:
            for revision, branch in self.conf.revisions:
                self.benchmark(revision, branch)

            for branch in self.conf.branches:
                self.benchmark(branch, branch)
        finally:
            self.report()
            if self.timings:
                self.report_timings()

        if self.failed:
            sys.exit(1)



import logging
import os.path
import sys


def cmd_list(options, benchmarks):
    print("%r benchmarks:" % options.benchmarks)
    for bench in sorted(benchmarks):
        print("- %s" % bench.name)
    print()
    print("Total: %s benchmarks" % len(benchmarks))


def cmd_list_groups(manifest, *, showtags=True):
    all_benchmarks = set(manifest.benchmarks)

    groups = sorted(manifest.groups - {"all", "default"})
    groups[0:0] = ["all", "default"]
    for group in groups:
        specs = list(manifest.resolve_group(group))
        known = set(specs) & all_benchmarks
        if not known:
            # skip empty groups
            continue

        print("%s (%s):" % (group, len(specs)))
        for spec in sorted(specs):
            print("- %s" % spec.name)
        print()

    if showtags:
        print("=============================")
        print()
        print("tags:")
        print()
        tags = sorted(manifest.tags or ())
        if not tags:
            print("(no tags)")
        else:
            for tag in tags:
                specs = list(manifest.resolve_group(tag))
                known = set(specs) & all_benchmarks
                if not known:
                    # skip empty groups
                    continue

                print("%s (%s):" % (tag, len(specs)))
                for spec in sorted(specs):
                    print("- %s" % spec.name)
                print()


def cmd_venv_create(options, root, python, benchmarks):
    from . import _venv
    from .venv import Requirements, VenvForBenchmarks

    if _venv.venv_exists(root):
        sys.exit(f"ERROR: the virtual environment already exists at {root}")

    requirements = Requirements.from_benchmarks(benchmarks)
    venv = VenvForBenchmarks.ensure(
        root,
        python or sys.executable,
        inherit_environ=options.inherit_environ,
    )
    venv.ensure_pip()
    try:
        venv.install_pyperformance()
        venv.ensure_reqs(requirements)
    except _venv.RequirementsInstallationFailedError:
        sys.exit(1)
    print("The virtual environment %s has been created" % root)


def cmd_venv_recreate(options, root, python, benchmarks):
    from . import _utils, _venv
    from .venv import Requirements, VenvForBenchmarks

    requirements = Requirements.from_benchmarks(benchmarks)
    if _venv.venv_exists(root):
        venv_python = _venv.resolve_venv_python(root)
        if venv_python == sys.executable:
            print("The virtual environment %s already exists" % root)
            print("(it matches the currently running Python executable)")
            venv = VenvForBenchmarks(
                root,
                inherit_environ=options.inherit_environ,
            )
            venv.ensure_pip()
            try:
                venv.ensure_reqs(requirements)
            except _venv.RequirementsInstallationFailedError:
                sys.exit(1)
        else:
            print("The virtual environment %s already exists" % root)
            _utils.safe_rmtree(root)
            print("The old virtual environment %s has been removed" % root)
            print()
            venv = VenvForBenchmarks.ensure(
                root,
                python or sys.executable,
                inherit_environ=options.inherit_environ,
            )
            venv.ensure_pip()
            try:
                venv.install_pyperformance()
                venv.ensure_reqs(requirements)
            except _venv.RequirementsInstallationFailedError:
                sys.exit(1)
            print("The virtual environment %s has been recreated" % root)
    else:
        venv = VenvForBenchmarks.create(
            root,
            python or sys.executable,
            inherit_environ=options.inherit_environ,
        )
        venv.ensure_pip()
        try:
            venv.install_pyperformance()
            venv.ensure_reqs(requirements)
        except _venv.RequirementsInstallationFailedError:
            sys.exit(1)
        print("The virtual environment %s has been created" % root)


def cmd_venv_remove(options, root):
    from . import _utils

    if _utils.safe_rmtree(root):
        print("The virtual environment %s has been removed" % root)
    else:
        print("The virtual environment %s does not exist" % root)


def cmd_venv_show(options, root):
    from . import _venv

    exists = _venv.venv_exists(root)

    text = "Virtual environment path: %s" % root
    if exists:
        text += " (already created)"
    else:
        text += " (not created yet)"
    print(text)

    if not exists:
        print()
        print("Command to create it:")
        cmd = "%s -m pyperformance venv create" % options.python
        if options.venv:
            cmd += " --venv=%s" % options.venv
        print(cmd)


def cmd_run(options, benchmarks):
    import pyperf

    import pyperformance

    from .compare import display_benchmark_suite
    from .run import run_benchmarks

    logging.basicConfig(level=logging.INFO)

    print("Python benchmark suite %s" % pyperformance.__version__)
    print()

    if options.output and os.path.exists(options.output):
        print("ERROR: the output file %s already exists!" % options.output)
        sys.exit(1)

    if hasattr(options, "python"):
        executable = options.python
    else:
        executable = sys.executable
    if not os.path.isabs(executable):
        print('ERROR: "%s" is not an absolute path' % executable)
        sys.exit(1)

    suite, errors = run_benchmarks(benchmarks, executable, options)

    if not suite:
        print("ERROR: No benchmark was run")
        sys.exit(1)

    if options.output:
        suite.dump(options.output)
    if options.append:
        pyperf.add_runs(options.append, suite)
    display_benchmark_suite(suite)

    if errors:
        print("%s benchmarks failed:" % len(errors))
        for name, reason in errors:
            print("- %s (%s)" % (name, reason))
        print()
        sys.exit(1)


def cmd_compile(options):
    from .compile import BenchmarkRevision, parse_config

    conf = parse_config(options.config_file, "compile")
    if options is not None:
        if options.no_update:
            conf.update = False
        if options.no_tune:
            conf.system_tune = False
    bench = BenchmarkRevision(
        conf, options.revision, options.branch, patch=options.patch, options=options
    )
    bench.main()


def cmd_compile_all(options):
    from .compile import BenchmarkAll

    bench = BenchmarkAll(options.config_file, options=options)
    bench.main()


def cmd_upload(options):
    import pyperf

    from .compile import BenchmarkRevision, parse_config, parse_date

    conf = parse_config(options.config_file, "upload")

    filename = options.json_file
    bench = pyperf.BenchmarkSuite.load(filename)
    metadata = bench.get_metadata()
    revision = metadata["commit_id"]
    branch = metadata["commit_branch"]
    commit_date = parse_date(metadata["commit_date"])

    bench = BenchmarkRevision(
        conf,
        revision,
        branch,
        filename=filename,
        commit_date=commit_date,
        setup_log=False,
        options=options,
    )
    bench.upload()


def cmd_show(options):
    import pyperf

    from .compare import display_benchmark_suite

    suite = pyperf.BenchmarkSuite.load(options.filename)
    display_benchmark_suite(suite)


def cmd_compare(options):
    from .compare import VersionMismatchError, compare_results, write_csv

    try:
        results = compare_results(options)
    except VersionMismatchError as exc:
        print(f"ERROR: {exc}")
        sys.exit(1)

    if options.csv:
        write_csv(results, options.csv)



"""
Benchmark recursive generators implemented in python
by traversing a binary tree.

Author: Kumar Aditya
"""

from __future__ import annotations

from collections.abc import Iterator

import pyperf


class Tree:
    def __init__(self, left: Tree | None, value: int, right: Tree | None) -> None:
        self.left = left
        self.value = value
        self.right = right

    def __iter__(self) -> Iterator[int]:
        if self.left:
            yield from self.left
        yield self.value
        if self.right:
            yield from self.right


def tree(input: range) -> Tree | None:
    n = len(input)
    if n == 0:
        return None
    i = n // 2
    return Tree(tree(input[:i]), input[i], tree(input[i + 1:]))

def bench_generators(loops: int) -> float:
    assert list(tree(range(10))) == list(range(10))
    range_it = range(loops)
    iterable = tree(range(100000))
    t0 = pyperf.perf_counter()
    for _ in range_it:
        for _ in iterable:
            pass
    return pyperf.perf_counter() - t0

if __name__ == "__main__":
    runner = pyperf.Runner()
    runner.metadata['description'] = "Benchmark generators"
    runner.bench_time_func('generators', bench_generators)



import csv
import math
import os.path
import statistics

import pyperf

NO_VERSION = "<not set>"


class VersionMismatchError(Exception):
    def __init__(self, version1, version2):
        super().__init__(
            f"Performance versions are different ({version1} != {version2})",
        )
        self.version1 = version1
        self.version2 = version2


def format_result(bench):
    mean = bench.mean()
    if bench.get_nvalue() >= 2:
        args = bench.format_values((mean, bench.stdev()))
        return "Mean +- std dev: %s +- %s" % args
    else:
        return bench.format_value(mean)


# A table of 95% confidence intervals for a two-tailed t distribution, as a
# function of the degrees of freedom. For larger degrees of freedom, we
# approximate. While this may look less elegant than simply calculating the
# critical value, those calculations suck. Look at
# http://www.math.unb.ca/~knight/utility/t-table.htm if you need more values.
_T_DIST_95_CONF_LEVELS = [
    0,
    12.706,
    4.303,
    3.182,
    2.776,
    2.571,
    2.447,
    2.365,
    2.306,
    2.262,
    2.228,
    2.201,
    2.179,
    2.160,
    2.145,
    2.131,
    2.120,
    2.110,
    2.101,
    2.093,
    2.086,
    2.080,
    2.074,
    2.069,
    2.064,
    2.060,
    2.056,
    2.052,
    2.048,
    2.045,
    2.042,
]


def tdist95conf_level(df):
    """Approximate the 95% confidence interval for Student's T distribution.

    Given the degrees of freedom, returns an approximation to the 95%
    confidence interval for the Student's T distribution.

    Args:
        df: An integer, the number of degrees of freedom.

    Returns:
        A float.
    """
    df = int(round(df))
    highest_table_df = len(_T_DIST_95_CONF_LEVELS)
    if df >= 200:
        return 1.960
    if df >= 100:
        return 1.984
    if df >= 80:
        return 1.990
    if df >= 60:
        return 2.000
    if df >= 50:
        return 2.009
    if df >= 40:
        return 2.021
    if df >= highest_table_df:
        return _T_DIST_95_CONF_LEVELS[highest_table_df - 1]
    return _T_DIST_95_CONF_LEVELS[df]


def pooled_sample_variance(sample1, sample2):
    """Find the pooled sample variance for two samples.

    Args:
        sample1: one sample.
        sample2: the other sample.

    Returns:
        Pooled sample variance, as a float.
    """
    deg_freedom = len(sample1) + len(sample2) - 2
    mean1 = statistics.mean(sample1)
    squares1 = ((x - mean1) ** 2 for x in sample1)
    mean2 = statistics.mean(sample2)
    squares2 = ((x - mean2) ** 2 for x in sample2)

    return (math.fsum(squares1) + math.fsum(squares2)) / float(deg_freedom)


def tscore(sample1, sample2):
    """Calculate a t-test score for the difference between two samples.

    Args:
        sample1: one sample.
        sample2: the other sample.

    Returns:
        The t-test score, as a float.
    """
    if len(sample1) != len(sample2):
        raise ValueError("different number of values")
    error = pooled_sample_variance(sample1, sample2) / len(sample1)
    diff = statistics.mean(sample1) - statistics.mean(sample2)
    return diff / math.sqrt(error * 2)


def is_significant(sample1, sample2):
    """Determine whether two samples differ significantly.

    This uses a Student's two-sample, two-tailed t-test with alpha=0.95.

    Args:
        sample1: one sample.
        sample2: the other sample.

    Returns:
        (significant, t_score) where significant is a bool indicating whether
        the two samples differ significantly; t_score is the score from the
        two-sample T test.
    """
    deg_freedom = len(sample1) + len(sample2) - 2
    critical_value = tdist95conf_level(deg_freedom)
    t_score = tscore(sample1, sample2)
    return (abs(t_score) >= critical_value, t_score)


def significant_msg(base, changed):
    if base.get_nvalue() < 2 or changed.get_nvalue() < 2:
        return "(benchmark only contains a single value)"

    avg_base = base.mean()
    avg_changed = changed.mean()

    msg = "Not significant"
    significant = False

    # Due to inherent measurement imprecisions, variations of less than 1%
    # are automatically considered insignificant. This helps present
    # a clear picture to the user.
    if abs(avg_base - avg_changed) > (avg_base + avg_changed) * 0.01:
        base_times = base.get_values()
        changed_times = changed.get_values()

        significant, t_score = is_significant(base_times, changed_times)
        if significant:
            msg = "Significant (t=%.2f)" % t_score

    return msg


def format_table(base_label, changed_label, results):
    table = [("Benchmark", base_label, changed_label, "Change", "Significance")]

    for bench_name, result in results:
        format_value = result.base.format_value
        avg_base = result.base.mean()
        avg_changed = result.changed.mean()
        delta_avg = quantity_delta(result.base, result.changed)
        msg = significant_msg(result.base, result.changed)
        table.append(
            (
                bench_name,
                # Limit the precision for conciseness in the table.
                format_value(avg_base),
                format_value(avg_changed),
                delta_avg,
                msg,
            )
        )

    # Columns with None values are skipped
    skipped_cols = set()
    col_widths = [0] * len(table[0])
    for row in table:
        for col, val in enumerate(row):
            if val is None:
                skipped_cols.add(col)
                continue
            col_widths[col] = max(col_widths[col], len(val))

    outside_line = "+"
    header_sep_line = "+"
    for col, width in enumerate(col_widths):
        if col in skipped_cols:
            continue
        width += 2  # Compensate for the left and right padding spaces.
        outside_line += "-" * width + "+"
        header_sep_line += "=" * width + "+"

    output = [outside_line]
    for row_i, row in enumerate(table):
        output_row = []
        for col_i, val in enumerate(row):
            if col_i in skipped_cols:
                continue
            output_row.append("| " + val.ljust(col_widths[col_i]) + " ")
        output.append("".join(output_row) + "|")
        if row_i > 0:
            output.append(outside_line)

    output.insert(2, "".join(header_sep_line))
    return "\n".join(output)


class BenchmarkResult(object):
    """An object representing data from a succesful benchmark run."""

    def __init__(self, base, changed):
        name = base.get_name()
        name2 = changed.get_name()
        if name2 != name:
            raise ValueError("not the same benchmark: %s != %s" % (name, name2))

        if base.get_nvalue() != changed.get_nvalue():
            raise RuntimeError("base and changed don't have the same number of values")

        self.base = base
        self.changed = changed

    def __str__(self):
        if self.base.get_nvalue() > 1:
            values = (
                self.base.mean(),
                self.base.stdev(),
                self.changed.mean(),
                self.changed.stdev(),
            )
            text = "%s +- %s -> %s +- %s" % self.base.format_values(values)

            msg = significant_msg(self.base, self.changed)
            delta_avg = quantity_delta(self.base, self.changed)
            return "Mean +- std dev: %s: %s\n%s" % (text, delta_avg, msg)
        else:
            format_value = self.base.format_value
            base = self.base.mean()
            changed = self.changed.mean()
            delta_avg = quantity_delta(self.base, self.changed)
            return "%s -> %s: %s" % (
                format_value(base),
                format_value(changed),
                delta_avg,
            )


def quantity_delta(base, changed):
    old = base.mean()
    new = changed.mean()
    is_time = base.get_unit() == "second"

    if old == 0 or new == 0:
        return "incomparable (one result was zero)"
    if new > old:
        if is_time:
            return "%.2fx slower" % (new / old)
        else:
            return "%.2fx larger" % (new / old)
    elif new < old:
        if is_time:
            return "%.2fx faster" % (old / new)
        else:
            return "%.2fx smaller" % (old / new)
    else:
        return "no change"


def display_suite_metadata(suite, title=None):
    metadata = suite.get_metadata()
    empty = True
    for key, fmt in (
        ("performance_version", "Performance version: %s"),
        ("python_version", "Python version: %s"),
        ("platform", "Report on %s"),
        ("cpu_count", "Number of logical CPUs: %s"),
    ):
        if key not in metadata:
            continue

        empty = False
        if title:
            print(title)
            print("=" * len(title))
            print()
            title = None

        text = fmt % metadata[key]
        print(text)

    dates = suite.get_dates()
    if dates:
        print("Start date: %s" % dates[0].isoformat(" "))
        print("End date: %s" % dates[1].isoformat(" "))
        empty = False

    if not empty:
        print()


def display_benchmark_suite(suite):
    display_suite_metadata(suite)

    for bench in suite.get_benchmarks():
        print("### %s ###" % bench.get_name())
        print(format_result(bench))
        print()


def get_labels(filename1, filename2):
    # Find a short label to identify two filenames:
    # the two labels must be different
    name1 = os.path.basename(filename1)
    name2 = os.path.basename(filename2)
    if name1 != name2:
        return (name1, name2)

    return (filename1, filename2)


def compare_results(options):
    base_label, changed_label = get_labels(
        options.baseline_filename, options.changed_filename
    )

    base_suite = pyperf.BenchmarkSuite.load(options.baseline_filename)
    changed_suite = pyperf.BenchmarkSuite.load(options.changed_filename)

    results = []
    common = set(base_suite.get_benchmark_names()) & set(
        changed_suite.get_benchmark_names()
    )
    for name in sorted(common):
        base_bench = base_suite.get_benchmark(name)
        changed_bench = changed_suite.get_benchmark(name)
        result = BenchmarkResult(base_bench, changed_bench)
        results.append(result)

    hidden = []
    shown = []
    for result in results:
        name = result.base.get_name()

        significant = significant_msg(result.base, result.changed)
        if significant or options.verbose:
            shown.append((name, result))
        else:
            hidden.append((name, result))

    display_suite_metadata(base_suite, title=base_label)
    display_suite_metadata(changed_suite, title=changed_label)

    if options.output_style == "normal":
        for index, item in enumerate(shown):
            if index:
                print()
            name, result = item
            print("###", name, "###")
            print(result)

    elif options.output_style == "table":
        if shown:
            print(format_table(base_label, changed_label, shown))
    else:
        raise ValueError("Invalid output_style: %r" % options.output_style)

    if hidden:
        print()
        print("The following not significant results are hidden, use -v to show them:")
        print(", ".join(name for (name, result) in hidden) + ".")

    only_base = set(base_suite.get_benchmark_names()) - common
    if only_base:
        print()
        print(
            "Skipped %s benchmarks only in %s: %s"
            % (len(only_base), base_label, ", ".join(sorted(only_base)))
        )

    only_changed = set(changed_suite.get_benchmark_names()) - common
    if only_changed:
        print()
        print(
            "Skipped %s benchmarks only in %s: %s"
            % (len(only_changed), changed_label, ", ".join(sorted(only_changed)))
        )

    version1 = base_suite.get_metadata().get("performance_version", NO_VERSION)
    version2 = changed_suite.get_metadata().get("performance_version", NO_VERSION)
    if version1 != version2 or (version1 == version2 == NO_VERSION):
        raise VersionMismatchError(version1, version2)

    return results


def format_csv(value):
    abs_value = abs(value)
    # keep at least 3 significant digits, but also try to avoid too many zeros
    if abs_value >= 1.0:
        return "%.2f" % value
    elif abs_value >= 1e-3:
        return "%.5f" % value
    elif abs_value >= 1e-6:
        return "%.8f" % value
    else:
        return "%.11f" % value


def write_csv(results, filename):
    with open(filename, "w", newline="", encoding="ascii") as fp:
        writer = csv.writer(fp)
        writer.writerow(["Benchmark", "Base", "Changed"])
        for result in results:
            name = result.base.get_name()
            base = result.base.mean()
            changed = result.changed.mean()
            row = [name, format_csv(base), format_csv(changed)]
            writer.writerow(row)
